from __future__ import annotations

import warnings
from functools import partial
from time import ctime
from typing import Union

import numpy as np
from ase.units import Ha
from ase.utils.timing import timer

import gpaw
import gpaw.mpi as mpi
from gpaw.bztools import convex_hull_volume
from gpaw.response.chi0_data import Chi0Data
from gpaw.response.frequencies import (FrequencyDescriptor,
                                       FrequencyGridDescriptor,
                                       NonLinearFrequencyDescriptor)
from gpaw.response.hilbert import HilbertTransform
from gpaw.response.integrators import (Integrator, PointIntegrator,
                                       TetrahedronIntegrator)
from gpaw.response.pair import NoCalculatorPairDensity
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.symmetry import PWSymmetryAnalyzer
from gpaw.typing import Array1D
from gpaw.utilities.memory import maxrss


def find_maximum_frequency(kpt_u, nbands=0, fd=None):
    """Determine the maximum electron-hole pair transition energy."""
    epsmin = 10000.0
    epsmax = -10000.0
    for kpt in kpt_u:
        epsmin = min(epsmin, kpt.eps_n[0])
        epsmax = max(epsmax, kpt.eps_n[nbands - 1])

    if fd is not None:
        print('Minimum eigenvalue: %10.3f eV' % (epsmin * Ha), file=fd)
        print('Maximum eigenvalue: %10.3f eV' % (epsmax * Ha), file=fd)

    return epsmax - epsmin


class Chi0Calculator:
    def __init__(self, wd, pair,
                 hilbert=True,
                 intraband=True,
                 nbands=None,
                 timeordered=False,
                 context=None,
                 ecut=None,
                 eta=0.2,
                 disable_point_group=False, disable_time_reversal=False,
                 disable_non_symmorphic=True,
                 integrationmode=None,
                 rate=0.0, eshift=0.0):

        if context is None:
            context = pair.context

        # TODO: More refactoring to avoid non-orthogonal inputs.
        assert pair.context.world is context.world
        self.context = context

        self.timer = self.context.timer
        self.fd = self.context.fd

        self.pair = pair
        self.gs = pair.gs

        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic
        self.integrationmode = integrationmode
        self.eshift = eshift / Ha

        self.vol = self.gs.volume
        self.world = self.context.world
        self.nblocks = pair.nblocks
        self.calc = self.gs._calc  # XXX remove me

        # XXX this is redundant as pair also does it.
        self.blockcomm, self.kncomm = block_partition(self.world, self.nblocks)

        if ecut is None:
            ecut = 50.0
        ecut /= Ha
        self.ecut = ecut

        self.eta = eta / Ha
        if rate == 'eta':
            self.rate = self.eta
        else:
            self.rate = rate / Ha

        self.nbands = nbands or self.gs.bd.nbands
        self.include_intraband = intraband

        self.wd = wd
        print(self.wd, file=self.fd)

        if not isinstance(self.wd, NonLinearFrequencyDescriptor):
            assert not hilbert

        self.hilbert = hilbert
        self.timeordered = bool(timeordered)

        if self.eta == 0.0:
            assert not hilbert
            assert not timeordered
            assert not self.wd.omega_w.real.any()

        self.nocc1 = self.pair.nocc1  # number of completely filled bands
        self.nocc2 = self.pair.nocc2  # number of non-empty bands

        self.Q_aGii = None

        if sum(self.pbc) == 1:
            raise ValueError('1-D not supported atm.')

        print('Nonperiodic BCs: ', (~self.pbc),
              file=self.fd)

        if integrationmode is not None:
            print('Using integration method: ' + self.integrationmode,
                  file=self.fd)
        else:
            print('Using integration method: PointIntegrator', file=self.fd)

    @property
    def pbc(self):
        return self.gs.pbc

    def create_chi0(self, q_c, extend_head=True):
        # Extract descriptor arguments
        plane_waves = (q_c, self.ecut, self.gs.gd)
        parallelization = (self.world, self.blockcomm, self.kncomm)

        # Construct the Chi0Data object
        # In the future, the frequencies should be specified at run-time
        # by Chi0.calculate(), in which case Chi0Data could also initialize
        # the frequency descriptor XXX
        chi0 = Chi0Data.from_descriptor_arguments(self.wd,
                                                  plane_waves,
                                                  parallelization,
                                                  extend_head)

        return chi0

    def calculate(self, q_c, spin='all'):
        """Calculate response function.

        Parameters
        ----------
        q_c : list or ndarray
            Momentum vector.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
            (not used in transverse response functions)

        Returns
        -------
        chi0 : Chi0Data
            Data object containing the chi0 data arrays along with basis
            representation descriptors and blocks distribution
        """
        gs = self.gs

        if spin == 'all':
            spins = range(gs.nspins)
        else:
            assert spin in range(gs.nspins)
            spins = [spin]

        chi0 = self.create_chi0(q_c)

        self.print_chi(chi0.pd)

        if chi0.optical_limit:
            self.plasmafreq_vv = np.zeros((3, 3), complex)
        else:
            self.plasmafreq_vv = None

        # Do all transitions into partially filled and empty bands
        m1 = self.nocc1
        m2 = self.nbands

        chi0 = self.update_chi0(chi0, m1, m2, spins)

        return chi0

    @timer('Calculate CHI_0')
    def update_chi0(self,
                    chi0: Chi0Data,
                    m1, m2, spins):
        """In-place calculation of the response function.

        Parameters
        ----------
        chi0 : Chi0Data
            Data and representation object
        m1 : int
            Lower band cutoff for band summation
        m2 : int
            Upper band cutoff for band summation
        spins : str or list(ints)
            If 'all' then include all spins.
            If [0] or [1], only include this specific spin.

        Returns
        -------
        chi0 : Chi0Data
        """
        assert m1 <= m2
        # Parse spins
        gs = self.gs

        if spins == 'all':
            spins = range(gs.nspins)
        else:
            for spin in spins:
                assert spin in range(gs.nspins)

        pd = chi0.pd
        # Are we calculating the optical limit.
        optical_limit = chi0.optical_limit

        # Use wings in optical limit, if head cannot be extended
        if optical_limit and not chi0.extend_head:
            wings = True
        else:
            wings = False

        # Reset PAW correction in case momentum has change
        self.Q_aGii = self.pair.initialize_paw_corrections(pd)
        A_wxx = chi0.chi0_wGG  # Change notation

        # Initialize integrator. The integrator class is a general class
        # for brillouin zone integration that can integrate user defined
        # functions over user defined domains and sum over bands.
        integrator: Integrator
        intnoblock: Integrator

        if self.integrationmode is None or \
           self.integrationmode == 'point integration':
            cls = PointIntegrator
        elif self.integrationmode == 'tetrahedron integration':
            cls = TetrahedronIntegrator  # type: ignore
        else:
            raise ValueError(f'Integration mode "{self.integrationmode}"'
                             ' not implemented.')

        kwargs = dict(
            cell_cv=self.gs.gd.cell_cv,
            comm=self.world,
            timer=self.timer,
            eshift=self.eshift,
            txt=self.fd)

        integrator = cls(**kwargs, nblocks=self.nblocks)
        intnoblock = cls(**kwargs)

        # The integration domain is determined by the following function
        # that reduces the integration domain to the irreducible zone
        # of the little group of q.
        bzk_kv, analyzer = self.get_kpoints(
            pd, integrationmode=self.integrationmode)
        domain = (bzk_kv, spins)

        if self.integrationmode == 'tetrahedron integration':
            # If there are non-periodic directions it is possible that the
            # integration domain is not compatible with the symmetry operations
            # which essentially means that too large domains will be
            # integrated. We normalize by vol(BZ) / vol(domain) to make
            # sure that to fix this.
            domainvol = convex_hull_volume(
                bzk_kv) * analyzer.how_many_symmetries()
            bzvol = (2 * np.pi)**3 / self.vol
            factor = bzvol / domainvol
        else:
            factor = 1

        prefactor = (2 * factor * analyzer.how_many_symmetries() /
                     (gs.nspins * (2 * np.pi)**3))  # Remember prefactor

        if self.integrationmode is None:
            nbzkpts = gs.kd.nbzkpts
            prefactor *= len(bzk_kv) / nbzkpts

        A_wxx /= prefactor
        if wings:
            chi0.chi0_wxvG /= prefactor
            chi0.chi0_wvv /= prefactor

        # The functions that are integrated are defined in the bottom
        # of this file and take a number of constant keyword arguments
        # which the integrator class accepts through the use of the
        # kwargs keyword.
        kd = gs.kd
        mat_kwargs = {'kd': kd, 'pd': pd,
                      'symmetry': analyzer,
                      'integrationmode': self.integrationmode}
        eig_kwargs = {'kd': kd, 'pd': pd}

        if not chi0.extend_head:
            mat_kwargs['extend_head'] = False

        # Determine what "kind" of integral to make.
        extraargs = {}  # Initialize extra arguments to integration method.
        if self.eta == 0:
            # If eta is 0 then we must be working with imaginary frequencies.
            # In this case chi is hermitian and it is therefore possible to
            # reduce the computational costs by a only computing half of the
            # response function.
            kind = 'hermitian response function'
        elif self.hilbert:
            # The spectral function integrator assumes that the form of the
            # integrand is a function (a matrix element) multiplied by
            # a delta function and should return a function of at user defined
            # x's (frequencies). Thus the integrand is tuple of two functions
            # and takes an additional argument (x).
            kind = 'spectral function'
        else:
            # Otherwise, we can make no simplifying assumptions of the
            # form of the response function and we simply perform a brute
            # force calculation of the response function.
            kind = 'response function'
            extraargs['eta'] = self.eta
            extraargs['timeordered'] = self.timeordered

        # Integrate response function
        print('Integrating response function.', file=self.fd)
        # Define band summation. Includes transitions from all
        # completely and partially filled bands to range(m1, m2)
        bandsum = {'n1': 0, 'n2': self.nocc2, 'm1': m1, 'm2': m2}
        mat_kwargs.update(bandsum)
        eig_kwargs.update(bandsum)

        integrator.integrate(kind=kind,  # Kind of integral
                             domain=domain,  # Integration domain
                             integrand=(self.get_matrix_element,  # Integrand
                                        self.get_eigenvalues),  # Integrand
                             x=self.wd,  # Frequency Descriptor
                             kwargs=(mat_kwargs, eig_kwargs),
                             # Arguments for integrand functions
                             out_wxx=A_wxx,  # Output array
                             **extraargs)
        # extraargs: Extra arguments to integration method
        if wings:
            mat_kwargs['extend_head'] = True
            mat_kwargs['block'] = False
            # This is horrible but we need to update the wings manually
            # in order to make them work with ralda, RPA and GW. This entire
            # section can be deleted in the future if the ralda and RPA code is
            # made compatible with the head and wing extension that other parts
            # of the code is using.
            chi0_wxvx = np.zeros(np.array(chi0.chi0_wxvG.shape) +
                                 [0, 0, 0, 2],
                                 complex)  # Notice the wxv"x" for head extend
            intnoblock.integrate(kind=kind + ' wings',  # kind'o int.
                                 domain=domain,  # Integration domain
                                 integrand=(self.get_matrix_element,  # Intgrnd
                                            self.get_eigenvalues),  # Integrand
                                 x=self.wd,  # Frequency Descriptor
                                 kwargs=(mat_kwargs, eig_kwargs),
                                 # Arguments for integrand functions
                                 out_wxx=chi0_wxvx,  # Output array
                                 **extraargs)

        if self.hilbert:
            # The integrator only returns the spectral function and a Hilbert
            # transform is performed to return the real part of the density
            # response function.
            with self.timer('Hilbert transform'):
                # Make Hilbert transform
                ht = HilbertTransform(np.array(self.wd.omega_w), self.eta,
                                      timeordered=self.timeordered)
                ht(A_wxx)
                if wings:
                    ht(chi0_wxvx)

        # In the optical limit additional work must be performed
        # for the intraband response.
        # Only compute the intraband response if there are partially
        # unoccupied bands and only if the user has not disabled its
        # calculation using the include_intraband keyword.
        if optical_limit and self.nocc1 != self.nocc2:
            # The intraband response is essentially just the calculation
            # of the free space Drude plasma frequency. The calculation is
            # similarly to the interband transitions documented above.
            mat_kwargs = {'kd': kd, 'symmetry': analyzer,
                          'n1': self.nocc1, 'n2': self.nocc2,
                          'pd': pd}  # Integrand arguments
            eig_kwargs = {'kd': kd,
                          'n1': self.nocc1, 'n2': self.nocc2,
                          'pd': pd}  # Integrand arguments
            domain = (bzk_kv, spins)  # Integration domain
            fermi_level = self.pair.fermi_level  # Fermi level

            # Not so elegant solution but it works
            plasmafreq_wvv = np.zeros((1, 3, 3), complex)  # Output array
            print('Integrating intraband density response.', file=self.fd)

            # Depending on which integration method is used we
            # have to pass different arguments
            extraargs = {}
            if self.integrationmode is None:
                # Calculate intraband transitions at finite fermi smearing
                extraargs['intraband'] = True  # Calculate intraband
            elif self.integrationmode == 'tetrahedron integration':
                # Calculate intraband transitions at T=0
                extraargs['x'] = FrequencyGridDescriptor([-fermi_level])

            intnoblock.integrate(kind='spectral function',  # Kind of integral
                                 domain=domain,  # Integration domain
                                 # Integrands
                                 integrand=(self.get_intraband_response,
                                            self.get_intraband_eigenvalue),
                                 # Integrand arguments
                                 kwargs=(mat_kwargs, eig_kwargs),
                                 out_wxx=plasmafreq_wvv,  # Output array
                                 **extraargs)  # Extra args for int. method

            # Again, not so pretty but that's how it is
            plasmafreq_vv = plasmafreq_wvv[0].copy()
            if self.include_intraband:
                drude_chi_wvv = plasmafreq_vv[np.newaxis]\
                    / (self.wd.omega_w[:, np.newaxis, np.newaxis]
                       + 1.j * self.rate)**2
                if chi0.extend_head:
                    va = min(chi0.blocks1d.a, 3)
                    vb = min(chi0.blocks1d.b, 3)
                    A_wxx[:, :vb - va, :3] += drude_chi_wvv[:, va:vb]
                else:
                    # Fill into head part of tmp head AND wings array
                    chi0_wxvx[:, 0, :3, :3] += drude_chi_wvv

            # Save the plasmafrequency
            try:
                self.plasmafreq_vv += 4 * np.pi * plasmafreq_vv * prefactor
            except AttributeError:
                self.plasmafreq_vv = 4 * np.pi * plasmafreq_vv * prefactor

            analyzer.symmetrize_wvv(self.plasmafreq_vv[np.newaxis])
            print('Plasma frequency:', file=self.fd)
            print((self.plasmafreq_vv**0.5 * Ha).round(2),
                  file=self.fd)

        # The response function is integrated only over the IBZ. The
        # chi calculated above must therefore be extended to include the
        # response from the full BZ. This extension can be performed as a
        # simple post processing of the response function that makes
        # sure that the response function fulfills the symmetries of the little
        # group of q. Due to the specific details of the implementation the chi
        # calculated above is normalized by the number of symmetries (as seen
        # below) and then symmetrized.
        A_wxx *= prefactor

        tmpA_wxx = chi0.blockdist.redistribute(A_wxx, chi0.nw)
        if chi0.extend_head:
            analyzer.symmetrize_wxx(tmpA_wxx,
                                    optical_limit=optical_limit)
        else:
            analyzer.symmetrize_wGG(tmpA_wxx)
            if wings:
                # Fill in wings part of the data, but leave out the head
                chi0.chi0_wxvG[..., 1:] += chi0_wxvx[..., 3:]
                # Fill in the head
                chi0.chi0_wvv += chi0_wxvx[:, 0, :3, :3]
                analyzer.symmetrize_wxvG(chi0.chi0_wxvG)
                analyzer.symmetrize_wvv(chi0.chi0_wvv)
        A_wxx[:] = chi0.blockdist.redistribute(tmpA_wxx, chi0.nw)

        # If point summation was used then the normalization of the
        # response function is not right and we have to make up for this
        # fact.

        if wings:
            chi0.chi0_wxvG *= prefactor
            chi0.chi0_wvv *= prefactor

        # In the optical limit, we have extended the wings and the head to
        # account for their nonanalytic behaviour which means that the size of
        # the chi0_wGG matrix is nw * (nG + 2)**2. Below we extract these
        # parameters.
        if optical_limit and chi0.extend_head:
            # We always return chi0 in the extend_head=False format. This
            # makes the update terminology inaccurate as we have to make a
            # new Chi0Data instance. In the future, extend_head=True should
            # be confined inside Chi0.update_chi0, so that the update
            # terminology is self-consistent
            chi0_new = self.create_chi0(pd.kd.bzk_kc[0], extend_head=False)

            # Make an extended wings object to temporarily hold the head AND
            # wings data
            chi0_wxvG = np.zeros(chi0.wxvG_shape, complex)

            # Extract the head and wings data. The x = 0 wing represents the
            # upper horizontal block, while the x = 1 wing represents the left
            # vertical block.
            # The data in A_wxx is distributed over "x"-rows, so we need to be
            # careful
            va = min(chi0.blocks1d.a, 3)  # Cartesian part of myslice
            vb = min(chi0.blocks1d.b, 3)
            # Fill in the x = 0 wing
            chi0_wxvG[:, 0, va:vb] = A_wxx[:, :vb - va]
            # Fill in the x = 1 wing
            chi0_wxvG[:, 1, :,
                      chi0.blocks1d.myslice] = np.transpose(
                A_wxx[..., :3], (0, 2, 1))

            # The head and wings are not distributed in the Chi0Data object,
            # so we collect the contributions from all blocks
            self.blockcomm.sum(chi0_wxvG)

            # Fill in the head
            # The x = 0 wing of the extended wings object has the "normal"
            # view of the head of chi0
            chi0_new.chi0_wvv[:] = chi0_wxvG[:, 0, :3, :3]
            # Fill in wings part of the data, but leave out the head
            chi0_new.chi0_wxvG[..., 1:] = chi0_wxvG[..., 3:]
            # Jesus, this is complicated

            # It is easiest to redistribute over freqs to pick body
            tmpA_wxx = chi0.blockdist.redistribute(A_wxx, chi0.nw)
            chi0_wGG = tmpA_wxx[:, 2:, 2:]
            chi0_new.chi0_wGG = chi0_new.blockdist.redistribute(chi0_wGG,
                                                                chi0.nw)

            # Rename
            chi0 = chi0_new

        elif optical_limit:
            # By default, we fill in the G=0 entries of chi0_wGG with the
            # wings evaluated along the z-direction.
            # The x = 1 wing represents the left vertical block, which is
            # distributed in chi0_wGG
            chi0.chi0_wGG[:, :, 0] = chi0.chi0_wxvG[:, 1, 2,
                                                    chi0.blocks1d.myslice]

            if self.blockcomm.rank == 0:  # rank with G=0 row
                # The x = 0 wing represents the upper horizontal block
                chi0.chi0_wGG[:, 0, :] = chi0.chi0_wxvG[:, 0, 2, :]
                chi0.chi0_wGG[:, 0, 0] = chi0.chi0_wvv[:, 2, 2]

        return chi0

    @timer('Get kpoints')
    def get_kpoints(self, pd, integrationmode=None):
        """Get the integration domain."""
        analyzer = PWSymmetryAnalyzer(
            self.gs.kd, pd,
            timer=self.timer, txt=self.fd,
            disable_point_group=self.disable_point_group,
            disable_time_reversal=self.disable_time_reversal,
            disable_non_symmorphic=self.disable_non_symmorphic)

        if integrationmode is None:
            K_gK = analyzer.group_kpoints()
            bzk_kc = np.array([self.gs.kd.bzk_kc[K_K[0]] for
                               K_K in K_gK])
        elif integrationmode == 'tetrahedron integration':
            bzk_kc = analyzer.get_reduced_kd(pbc_c=self.pbc).bzk_kc
            if (~self.pbc).any():
                bzk_kc = np.append(bzk_kc,
                                   bzk_kc + (~self.pbc).astype(int),
                                   axis=0)

        bzk_kv = np.dot(bzk_kc, pd.gd.icell_cv) * 2 * np.pi

        return bzk_kv, analyzer

    @timer('Get matrix element')
    def get_matrix_element(self, k_v, s, n1=None, n2=None,
                           m1=None, m2=None,
                           pd=None, kd=None,
                           symmetry=None, integrationmode=None,
                           extend_head=True, block=True):
        """A function that returns pair-densities.

        A pair density is defined as::

         <snk| e^(-i (q + G) r) |s'mk+q>,

        where s and s' are spins, n and m are band indices, k is
        the kpoint and q is the momentum transfer. For dielectric
        response s'=s, for the transverse magnetic response
        s' is flipped with respect to s.

        Parameters
        ----------
        k_v : ndarray
            Kpoint coordinate in cartesian coordinates.
        s : int
            Spin index.
        n1 : int
            Lower occupied band index.
        n2 : int
            Upper occupied band index.
        m1 : int
            Lower unoccupied band index.
        m2 : int
            Upper unoccupied band index.
        pd : PlanewaveDescriptor instance
        kd : KpointDescriptor instance
            Calculator kpoint descriptor.
        symmetry: gpaw.response.pair.PWSymmetryAnalyzer instance
            Symmetry analyzer object for handling symmetries of the kpoints.
        integrationmode : str
            The integration mode employed.
        extend_head: Bool
            Extend the head to include non-analytic behaviour

        Return
        ------
        n_nmG : ndarray
            Pair densities.
        """
        assert m1 <= m2

        k_c = np.dot(pd.gd.cell_cv, k_v) / (2 * np.pi)

        q_c = pd.kd.bzk_kc[0]

        optical_limit = np.allclose(q_c, 0.0)

        nG = pd.ngmax
        weight = np.sqrt(symmetry.get_kpoint_weight(k_c) /
                         symmetry.how_many_symmetries())
        if self.Q_aGii is None:
            self.Q_aGii = self.pair.initialize_paw_corrections(pd)

        kptpair = self.pair.get_kpoint_pair(pd, s, k_c, n1, n2,
                                            m1, m2, block=block)

        m_m = np.arange(m1, m2)
        n_n = np.arange(n1, n2)

        n_nmG = self.pair.get_pair_density(pd, kptpair, n_n, m_m,
                                           Q_aGii=self.Q_aGii, block=block)

        if integrationmode is None:
            n_nmG *= weight

        df_nm = kptpair.get_occupation_differences(n_n, m_m)
        df_nm[df_nm <= 1e-20] = 0.0
        n_nmG *= df_nm[..., np.newaxis]**0.5

        if not extend_head and optical_limit:
            n_nmG = np.copy(n_nmG[:, :, 2:])
            optical_limit = False

        if extend_head and optical_limit:
            return n_nmG.reshape(-1, nG + 2 * optical_limit)
        else:
            return n_nmG.reshape(-1, nG)

    @timer('Get eigenvalues')
    def get_eigenvalues(self, k_v, s, n1=None, n2=None,
                        m1=None, m2=None,
                        kd=None, pd=None, gs=None,
                        filter=False):
        """A function that can return the eigenvalues.

        A simple function describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""
        if gs is None:
            gs = self.gs

        kd = gs.kd
        k_c = np.dot(pd.gd.cell_cv, k_v) / (2 * np.pi)
        q_c = pd.kd.bzk_kc[0]
        K1 = self.pair.find_kpoint(k_c)
        K2 = self.pair.find_kpoint(k_c + q_c)

        ik1 = kd.bz2ibz_k[K1]
        ik2 = kd.bz2ibz_k[K2]
        kpt1 = gs.kpt_qs[ik1][s]
        assert gs.kd.comm.size == 1
        kpt2 = gs.kpt_qs[ik2][s]
        deps_nm = np.subtract(kpt1.eps_n[n1:n2][:, np.newaxis],
                              kpt2.eps_n[m1:m2])

        if filter:
            fermi_level = self.pair.fermi_level
            deps_nm[kpt1.eps_n[n1:n2] > fermi_level, :] = np.nan
            deps_nm[:, kpt2.eps_n[m1:m2] < fermi_level] = np.nan

        return deps_nm.reshape(-1)

    def get_intraband_response(self, k_v, s, n1=None, n2=None,
                               kd=None, symmetry=None, pd=None,
                               integrationmode=None):
        k_c = np.dot(pd.gd.cell_cv, k_v) / (2 * np.pi)
        kpt1 = self.pair.get_k_point(s, k_c, n1, n2)
        n_n = range(n1, n2)

        vel_nv = self.pair.intraband_pair_density(kpt1, n_n)

        if self.integrationmode is None:
            f_n = kpt1.f_n
            width = self.gs.get_occupations_width()
            if width > 1e-15:
                dfde_n = - 1. / width * (f_n - f_n**2.0)
            else:
                dfde_n = np.zeros_like(f_n)
            vel_nv *= np.sqrt(-dfde_n[:, np.newaxis])
            weight = np.sqrt(symmetry.get_kpoint_weight(k_c) /
                             symmetry.how_many_symmetries())
            vel_nv *= weight

        return vel_nv

    @timer('Intraband eigenvalue')
    def get_intraband_eigenvalue(self, k_v, s,
                                 n1=None, n2=None, kd=None, pd=None):
        """A function that can return the eigenvalues.

        A simple function describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""
        gs = self.gs
        kd = gs.kd
        k_c = np.dot(pd.gd.cell_cv, k_v) / (2 * np.pi)
        K1 = self.pair.find_kpoint(k_c)
        ik = kd.bz2ibz_k[K1]
        kpt1 = gs.kpt_qs[ik][s]
        assert gs.kd.comm.size == 1

        return kpt1.eps_n[n1:n2]

    def print_chi(self, pd):
        gs = self.gs
        gd = gs.gd

        if gpaw.dry_run:
            from gpaw.mpi import SerialCommunicator
            size = gpaw.dry_run
            world = SerialCommunicator()
            world.size = size
        else:
            world = self.world

        q_c = pd.kd.bzk_kc[0]
        nw = len(self.wd)
        ecut = self.ecut * Ha
        ns = gs.nspins
        nbands = self.nbands
        nk = gs.kd.nbzkpts
        nik = gs.kd.nibzkpts
        ngmax = pd.ngmax
        eta = self.eta * Ha
        wsize = world.size
        knsize = self.kncomm.size
        nocc = self.nocc1
        npocc = self.nocc2
        ngridpoints = gd.N_c[0] * gd.N_c[1] * gd.N_c[2]
        nstat = (ns * npocc + world.size - 1) // world.size
        occsize = nstat * ngridpoints * 16. / 1024**2
        bsize = self.blockcomm.size
        chisize = nw * pd.ngmax**2 * 16. / 1024**2 / bsize

        p = partial(print, file=self.fd)

        p('%s' % ctime())
        p('Called response.chi0.calculate with')
        p('    q_c: [%f, %f, %f]' % (q_c[0], q_c[1], q_c[2]))
        p('    Number of frequency points: %d' % nw)
        if bsize > nw:
            p('WARNING! Your nblocks is larger than number of frequency'
              ' points. Errors might occur, if your submodule does'
              ' not know how to handle this.')
        p('    Planewave cutoff: %f' % ecut)
        p('    Number of spins: %d' % ns)
        p('    Number of bands: %d' % nbands)
        p('    Number of kpoints: %d' % nk)
        p('    Number of irredicible kpoints: %d' % nik)
        p('    Number of planewaves: %d' % ngmax)
        p('    Broadening (eta): %f' % eta)
        p('    world.size: %d' % wsize)
        p('    kncomm.size: %d' % knsize)
        p('    blockcomm.size: %d' % bsize)
        p('    Number of completely occupied states: %d' % nocc)
        p('    Number of partially occupied states: %d' % npocc)
        p()
        p('    Memory estimate of potentially large arrays:')
        p('        chi0_wGG: %f M / cpu' % chisize)
        p('        Occupied states: %f M / cpu' % occsize)
        p('        Memory usage before allocation: %f M / cpu' % (maxrss() /
                                                                  1024**2))
        p()


class Chi0(Chi0Calculator):
    """Class for calculating non-interacting response functions."""

    def __init__(self,
                 calc,
                 *,
                 frequencies: Union[dict, Array1D] = None,
                 ecut=50,
                 ftol=1e-6, threshold=1,
                 real_space_derivatives=False,
                 world=mpi.world, txt='-', timer=None,
                 nblocks=1,
                 nbands=None,
                 domega0=None,  # deprecated
                 omega2=None,  # deprecated
                 omegamax=None,  # deprecated
                 **kwargs):
        """Construct Chi0 object.

        Parameters
        ----------
        calc : str
            The groundstate calculation file that the linear response
            calculation is based on.
        frequencies :
            Input parameters for frequency_grid.
            Can be array of frequencies to evaluate the response function at
            or dictionary of paramaters for build-in nonlinear grid
            (see :ref:`frequency grid`).
        ecut : float
            Energy cutoff.
        hilbert : bool
            Switch for hilbert transform. If True, the full density response
            is determined from a hilbert transform of its spectral function.
            This is typically much faster, but does not work for imaginary
            frequencies.
        nbands : int
            Maximum band index to include.
        timeordered : bool
            Switch for calculating the time ordered density response function.
            In this case the hilbert transform cannot be used.
        eta : float
            Artificial broadening of spectra.
        ftol : float
            Threshold determining whether a band is completely filled
            (f > 1 - ftol) or completely empty (f < ftol).
        threshold : float
            Numerical threshold for the optical limit k dot p perturbation
            theory expansion (used in gpaw/response/pair.py).
        real_space_derivatives : bool
            Switch for calculating nabla matrix elements (in the optical limit)
            using a real space finite difference approximation.
        intraband : bool
            Switch for including the intraband contribution to the density
            response function.
        world : MPI comm instance
            MPI communicator.
        txt : str
            Output file.
        timer : gpaw.utilities.timing.timer instance
        nblocks : int
            Divide the response function into nblocks. Useful when the response
            function is large.
        disable_point_group : bool
            Do not use the point group symmetry operators.
        disable_time_reversal : bool
            Do not use time reversal symmetry.
        disable_non_symmorphic : bool
            Do no use non symmorphic symmetry operators.
        integrationmode : str
            Integrator for the kpoint integration.
            If == 'tetrahedron integration' then the kpoint integral is
            performed using the linear tetrahedron method.
        eshift : float
            Shift unoccupied bands
        rate : float,str
            Phenomenological scattering rate to use in optical limit Drude term
            (in eV). If rate='eta', then use input artificial broadening eta as
            rate. Note, for consistency with the formalism the rate is
            implemented as omegap^2 / (omega + 1j * rate)^2 which differ from
            some literature by a factor of 2.


        Attributes
        ----------
        pair : gpaw.response.pair.PairDensity instance
            Class for calculating matrix elements of pairs of wavefunctions.

        """
        from gpaw.response.context import calc_and_context
        calc, context = calc_and_context(calc, txt, world, timer)
        gs = calc.gs_adapter()
        nbands = nbands or gs.bd.nbands

        wd = new_frequency_descriptor(gs, nbands, frequencies, fd=context.fd,
                                      domega0=domega0,
                                      omega2=omega2, omegamax=omegamax)

        pair = NoCalculatorPairDensity(
            gs=gs, ftol=ftol, threshold=threshold,
            real_space_derivatives=real_space_derivatives,
            context=context,
            nblocks=nblocks)

        super().__init__(wd=wd, pair=pair, nbands=nbands, ecut=ecut, **kwargs)


def new_frequency_descriptor(gs, nbands, frequencies=None, *, fd,
                             domega0=None, omega2=None, omegamax=None):
    if domega0 is not None or omega2 is not None or omegamax is not None:
        assert frequencies is None
        frequencies = {'type': 'nonlinear',
                       'domega0': domega0,
                       'omega2': omega2,
                       'omegamax': omegamax}
        warnings.warn(f'Please use frequencies={frequencies}')

    elif frequencies is None:
        frequencies = {'type': 'nonlinear'}

    if (isinstance(frequencies, dict) and
        frequencies.get('omegamax') is None):
        omegamax = find_maximum_frequency(gs.kpt_u,
                                          nbands=nbands,
                                          fd=fd)
        frequencies['omegamax'] = omegamax * Ha

    wd = FrequencyDescriptor.from_array_or_dict(frequencies)
    return wd



from ase.optimize import BFGS
from ase.build import molecule
from ase.parallel import paropen
from gpaw import GPAW, PW
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy

# N
N = molecule('N')
N.cell = (6, 6, 7)
N.center()
calc = GPAW(mode=PW(600, force_complex_dtype=True),
            nbands=16,
            maxiter=300,
            xc='PBE',
            hund=True,
            txt='N_pbe.txt',
            parallel={'domain': 1},
            convergence={'density': 1.e-6})

N.calc = calc
E1_pbe = N.get_potential_energy()

calc.write('N.gpw', mode='all')

E1_hf = nsc_energy('N.gpw', 'EXX')

calc.diagonalize_full_hamiltonian(nbands=4800)
calc.write('N.gpw', mode='all')

# N2
N2 = molecule('N2')
N2.cell = (6, 6, 7)
N2.center()
calc = GPAW(mode=PW(600, force_complex_dtype=True),
            nbands=16,
            maxiter=300,
            xc='PBE',
            txt='N2_pbe.txt',
            parallel={'domain': 1},
            convergence={'density': 1.e-6})

N2.calc = calc
dyn = BFGS(N2)
dyn.run(fmax=0.05)
E2_pbe = N2.get_potential_energy()

calc.write('N2.gpw', mode='all')

E2_hf = nsc_energy('N2.gpw', 'EXX')

with paropen('PBE_HF.dat', 'w') as fd:
    print('PBE: ', E2_pbe - 2 * E1_pbe, file=fd)
    print('HF: ', E2_hf - 2 * E1_hf, file=fd)

calc.diagonalize_full_hamiltonian(nbands=4800)
calc.write('N2.gpw', mode='all')



"""
Calculate the magnetic response in iron using ALDA.

Tests whether the magnon energies and scattering intensities
have changed for:
 * Different kernel calculation strategies
 * Different chi0 transitions summation strategies
"""

import numpy as np
import pytest

import time

from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.response.tms import TransverseMagneticSusceptibility
from gpaw.response.susceptibility import read_macroscopic_component
from gpaw.test import findpeak, equal
from gpaw.mpi import world


pytestmark = pytest.mark.skipif(world.size < 4, reason='world.size < 4')


@pytest.mark.kspair
@pytest.mark.response
def test_response_iron_sf_ALDA(in_tmp_dir, scalapack):
    # ------------------- Inputs ------------------- #

    # Part 1: ground state calculation
    xc = 'LDA'
    kpts = 4
    nb = 6
    pw = 300
    conv = {'density': 1.e-8,
            'forces': 1.e-8}
    a = 2.867
    mm = 2.21

    # Part 2: magnetic response calculation
    q_c = [0.0, 0.0, 1 / 4.]
    fxc = 'ALDA'
    ecut = 300
    eta = 0.01

    # Test different kernel, summation and symmetry strategies
    # rshelmax, rshewmin, bandsummation, bundle_integrals, disable_syms
    strat_sd = [(None, None, 'pairwise', True, False),
                (-1, 0.001, 'pairwise', True, False),
                (-1, 0.001, 'pairwise', False, False),
                (-1, 0.000001, 'pairwise', True, False),
                (-1, 0.000001, 'double', True, False),
                (-1, 0.000001, 'double', True, True),
                (-1, None, 'pairwise', True, False),
                (3, None, 'pairwise', True, False)]
    frq_sw = [np.linspace(0.160, 0.320, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21),
              np.linspace(0.320, 0.480, 21)]

    # ------------------- Script ------------------- #

    # Part 1: ground state calculation

    t1 = time.time()

    Febcc = bulk('Fe', 'bcc', a=a)
    Febcc.set_initial_magnetic_moments([mm])

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts=monkhorst_pack((kpts, kpts, kpts)),
                nbands=nb,
                convergence=conv,
                symmetry={'point_group': False},
                parallel={'domain': 1})

    Febcc.calc = calc
    Febcc.get_potential_energy()
    t2 = time.time()

    # Part 2: magnetic response calculation

    for s, ((rshelmax, rshewmin, bandsummation, bundle_integrals,
             disable_syms), frq_w) in enumerate(zip(strat_sd, frq_sw)):
        tms = TransverseMagneticSusceptibility(
            calc,
            fxc=fxc,
            eta=eta,
            ecut=ecut,
            bandsummation=bandsummation,
            fxckwargs={'rshelmax': rshelmax,
                       'rshewmin': rshewmin},
            bundle_integrals=bundle_integrals,
            disable_point_group=disable_syms,
            disable_time_reversal=disable_syms,
            nblocks=2)
        tms.get_macroscopic_component(
            '+-', q_c, frq_w,
            filename='iron_dsus' + '_G%d.csv' % (s + 1))
        tms.write_timer()

    t3 = time.time()

    parprint('Ground state calculation took', (t2 - t1) / 60, 'minutes')
    parprint('Excited state calculations took', (t3 - t2) / 60, 'minutes')

    world.barrier()

    # Part 3: identify magnon peak in scattering functions
    w1_w, chiks1_w, chi1_w = read_macroscopic_component('iron_dsus_G1.csv')
    w2_w, chiks2_w, chi2_w = read_macroscopic_component('iron_dsus_G2.csv')
    w3_w, chiks3_w, chi3_w = read_macroscopic_component('iron_dsus_G3.csv')
    w4_w, chiks4_w, chi4_w = read_macroscopic_component('iron_dsus_G4.csv')
    w5_w, chiks5_w, chi5_w = read_macroscopic_component('iron_dsus_G5.csv')
    w6_w, chiks6_w, chi6_w = read_macroscopic_component('iron_dsus_G6.csv')
    w7_w, chiks7_w, chi7_w = read_macroscopic_component('iron_dsus_G7.csv')
    w8_w, chiks8_w, chi8_w = read_macroscopic_component('iron_dsus_G8.csv')

    wpeak1, Ipeak1 = findpeak(w1_w, -chi1_w.imag)
    wpeak2, Ipeak2 = findpeak(w2_w, -chi2_w.imag)
    wpeak3, Ipeak3 = findpeak(w3_w, -chi3_w.imag)
    wpeak4, Ipeak4 = findpeak(w4_w, -chi4_w.imag)
    wpeak5, Ipeak5 = findpeak(w5_w, -chi5_w.imag)
    wpeak6, Ipeak6 = findpeak(w6_w, -chi6_w.imag)
    wpeak7, Ipeak7 = findpeak(w7_w, -chi7_w.imag)
    wpeak8, Ipeak8 = findpeak(w8_w, -chi8_w.imag)

    mw1 = wpeak1 * 1000
    mw2 = wpeak2 * 1000
    mw3 = wpeak3 * 1000
    mw4 = wpeak4 * 1000
    mw5 = wpeak5 * 1000
    mw6 = wpeak6 * 1000
    mw7 = wpeak7 * 1000
    mw8 = wpeak8 * 1000

    # Part 4: compare new results to test values
    test_mw1 = 234.63  # meV
    test_mw2 = 397.33  # meV
    test_mw4 = 398.83  # meV
    test_Ipeak1 = 56.74  # a.u.
    test_Ipeak2 = 55.80  # a.u.
    test_Ipeak4 = 58.23  # a.u.

    # Different kernel strategies should remain the same
    # Magnon peak:
    equal(mw1, test_mw1, eta * 250)
    equal(mw2, test_mw2, eta * 250)
    equal(mw4, test_mw4, eta * 250)

    # Scattering function intensity:
    equal(Ipeak1, test_Ipeak1, 2.5)
    equal(Ipeak2, test_Ipeak2, 2.5)
    equal(Ipeak4, test_Ipeak4, 2.5)

    # The bundled and unbundled integration methods should give the same
    equal(mw2, mw3, eta * 100)
    equal(Ipeak2, Ipeak3, 1.0)

    # The two transitions summation strategies should give identical results
    equal(mw4, mw5, eta * 100)
    equal(Ipeak4, Ipeak5, 1.0)

    # Toggling symmetry should preserve the result
    equal(mw6, mw4, eta * 100)
    equal(Ipeak6, Ipeak4, 1.0)

    # Including vanishing coefficients should not matter for the result
    equal(mw7, mw4, eta * 100)
    equal(Ipeak7, Ipeak4, 1.0)
    equal(mw8, mw2, eta * 100)
    equal(Ipeak8, Ipeak2, 1.0)



import pickle

import numpy as np
from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw import PW
from gpaw.response.g0w0 import G0W0


a = 3.567
atoms = bulk('C', 'diamond', a=a)

k = 8

calc = GPAW(mode=PW(600, force_complex_dtype=True),
            parallel={'domain': 1},
            kpts={'size': (k, k, k), 'gamma': True},
            dtype=complex,
            xc='LDA',
            occupations=FermiDirac(0.001),
            txt='C_converged_ppa.txt')

atoms.calc = calc
atoms.get_potential_energy()

calc.diagonalize_full_hamiltonian()
calc.write('C_converged_ppa.gpw', 'all')

for ecut in [300, 400]:
    gw = G0W0(calc='C_converged_ppa.gpw',
              kpts=[0],
              bands=(3, 5),
              ecut=ecut,
              ppa=True,
              filename=f'C-g0w0_ppa_{ecut}')

    gw.calculate()

fil = pickle.load(open('C-g0w0_ppa_300_results_GW.pckl', 'rb'))
direct_gap_300 = fil['qp'][0, 0, 1] - fil['qp'][0, 0, 0]
fil = pickle.load(open('C-g0w0_ppa_400_results_GW.pckl', 'rb'))
direct_gap_400 = fil['qp'][0, 0, 1] - fil['qp'][0, 0, 0]

extrap_gap, slope = np.linalg.solve(np.array([[1, 1. / 300.**(3. / 2)],
                                              [1, 1. / 400.**(3. / 2)]]),
                                    np.array([direct_gap_300, direct_gap_400]))
print('Direct gap:', extrap_gap)



# web-page: mac_eps.csv
# Refer to G. Kresse, Phys. Rev. B 73, 045112 (2006)
# for comparison of macroscopic and microscopic dielectric constant
# and absorption peaks.
from pathlib import Path

from ase.build import bulk
from ase.parallel import paropen, world

from gpaw import GPAW, FermiDirac
from gpaw.response.df import DielectricFunction

# Ground state calculation
a = 5.431
atoms = bulk('Si', 'diamond', a=a)

calc = GPAW(mode='pw',
            kpts={'density': 5.0, 'gamma': True},
            parallel={'band': 1, 'domain': 1},
            xc='LDA',
            occupations=FermiDirac(0.001))  # use small FD smearing

atoms.calc = calc
atoms.get_potential_energy()  # get ground state density

# Restart Calculation with fixed density and dense kpoint sampling
calc = calc.fixed_density(
    kpts={'density': 15.0, 'gamma': False})  # dense kpoint sampling

calc.diagonalize_full_hamiltonian(nbands=70)  # diagonalize Hamiltonian
calc.write('si_large.gpw', 'all')  # write wavefunctions

# Getting absorption spectrum
df = DielectricFunction(calc='si_large.gpw',
                        eta=0.05,
                        domega0=0.02,
                        ecut=150)
df.get_dielectric_function(filename='si_abs.csv')

# Getting macroscopic constant
df = DielectricFunction(calc='si_large.gpw',
                        frequencies=[0.0],
                        hilbert=False,
                        eta=0.0001,
                        ecut=150,
                        )

epsNLF, epsLF = df.get_macroscopic_dielectric_constant()

# Make table
epsrefNLF = 14.08  # from [1] in top
epsrefLF = 12.66  # from [1] in top

with paropen('mac_eps.csv', 'w') as f:
    print(' , Without LFE, With LFE', file=f)
    print(f"{'GPAW-linear response'}, {epsNLF:.6f}, {epsLF:.6f}", file=f)
    print(f"{'[1]'}, {epsrefNLF:.6f}, {epsrefLF:.6f}", file=f)
    print(f"{'Exp.'}, {11.9:.6f}, {11.9:.6f}", file=f)

if world.rank == 0:
    Path('si_large.gpw').unlink()



# web-page: optimization.txt

from gpaw import restart
from ase.parallel import paropen as open
from ase.optimize import QuasiNewton


molecule, calc = restart('H2.gpw', txt='H2-relaxed.txt')

e2 = molecule.get_potential_energy()
d0 = molecule.get_distance(0, 1)

with open('optimization.txt', 'w') as fd:
    print('experimental bond length:', file=fd)
    print(f'hydrogen molecule energy: {e2:5.2f} eV', file=fd)
    print(f'bondlength              : {d0:5.2f} Ang', file=fd)
    
    # Find the theoretical bond length:
    relax = QuasiNewton(molecule, logfile='qn.log')
    relax.run(fmax=0.05)
    
    e2 = molecule.get_potential_energy()
    d0 = molecule.get_distance(0, 1)
    
    print(file=fd)
    print('PBE energy minimum:', file=fd)
    print(f'hydrogen molecule energy: {e2:5.2f} eV', file=fd)
    print(f'bondlength              : {d0:5.2f} Ang', file=fd)



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
## Introduction to Nudged Elastic Band (NEB) calculations

This tutorial describes how to use the NEB method to calculate the diffusion
barrier for an Au atom on Al(001). If you are not familiar with the NEB
method some relevant references are listed
[here.](https://wiki.fysik.dtu.dk/ase/ase/neb.html)

The tutorial uses the EMT potential in stead of DFT, as this is a lot faster.
It is based on a [tutorial found on the ASE
webpage](https://wiki.fysik.dtu.dk/ase/tutorials/neb/diffusion.html#diffusion-tutorial).

"""

# %%
# magic: %matplotlib notebook
import matplotlib.pyplot as plt
from ase.build import fcc100, add_adsorbate
from ase.constraints import FixAtoms
from ase.calculators.emt import EMT
from ase.optimize import BFGS
from ase.visualize import view
from ase.neb import NEB
from ase.io import read

# %%
"""
First we set up the initial state and check that it looks ok:
"""

# %%
# 2x2-Al(001) surface with 3 layers and an
# Au atom adsorbed in a hollow site:
slab = fcc100('Al', size=(2, 2, 3))
add_adsorbate(slab, 'Au', 1.7, 'hollow')
slab.center(axis=2, vacuum=4.0)
view(slab)

# %%
"""
Then we optimise the structure and save it
"""

# %%
# Fix second and third layers:
mask = [atom.tag > 1 for atom in slab]
slab.set_constraint(FixAtoms(mask=mask))

# Use EMT potential:
slab.calc = EMT()

# Optimise initial state:
qn = BFGS(slab, trajectory='initial.traj')
qn.run(fmax=0.05)

# %%
"""
We make the final state by moving the Au atom one lattice constant and
optimise again.
"""

# %%
# Optimise final state:
slab[-1].x += slab.get_cell()[0, 0] / 2
qn = BFGS(slab, trajectory='final.traj')
qn.run(fmax=0.05)

# %%
"""
Now we make a NEB calculation with 3 images between the inital and final
states. The images are initially made as copies of the initial state and the
command `interpolate()` makes a linear interpolation between the initial and
final state. As always, we check that everything looks ok before we run the
calculation.

NOTE: The linear interpolation works well in this case but not for e.g.
rotations. In this case an improved starting guess can be made with the [IDPP
method.](https://wiki.fysik.dtu.dk/ase/tutorials/neb/idpp.html#idpp-tutorial)
"""

# %%
initial = read('initial.traj')
final = read('final.traj')

constraint = FixAtoms(mask=[atom.tag > 1 for atom in initial])

n_im = 3  # number of images
images = [initial]
for i in range(n_im):
    image = initial.copy()
    image.calc = EMT()
    image.set_constraint(constraint)
    images.append(image)

images.append(final)

neb = NEB(images)
neb.interpolate()
view(images)

# %%
qn = BFGS(neb, trajectory='neb.traj')
qn.run(fmax=0.05)

# %%
"""
We visualize the final path with:
"""

# %%
view(images)

# %%
"""
You can find the barrier by selecting Tools->NEB in the gui (unfortunately,
the gui cannot show graphs when started from a notebook), or you can make a
script using
[NEBTools](https://wiki.fysik.dtu.dk/ase/ase/neb.html#ase.neb.NEBTools),
e.g.:
"""

# %%
from ase.neb import NEBTools

images = read('neb.traj@-5:')

nebtools = NEBTools(images)

# Get the calculated barrier and the energy change of the reaction.
Ef, dE = nebtools.get_barrier()
print('Barrier:', Ef)
print('Reaction energy:', dE)

# Generate new matplotlib axis - otherwise nebtools plot double.
fig, ax = plt.subplots()
nebtools.plot_band(ax)

# %%
"""
## Exercise

Now you should make your own NEB using the configuration with N<sub>2</sub>
lying down as the initial state and the configuration with two N atoms
adsorbed on the surface as the final state. The NEB needs to run in parallel
so you should make it as a python script, however you can use the Notebook to
test your configurations (but not the parallelisation) if you like and export
it as a script in the end.

### Parallelisation

The NEB should be parallelised over images. An example can be found in [this
GPAW tutorial](https://wiki.fysik.dtu.dk/gpaw/tutorialsexercises/moleculardynamics/neb/neb.html). The
script enumerates the cpu's and uses this number (``rank``) along with the
total number of cpu's (``size``) to distribute the tasks.
"""

# %%
# This code is just for illustration
from ase.parallel import world
n_im = 4              # Number of images
n = world.size // n_im      # number of cpu's per image
j = 1 + world.rank // n     # image number on this cpu
assert n_im * n == world.size

# %%
"""
For each image we assign a set of cpu's identified by their rank. The rank
numbers are given to the calculator associated with this image.
"""

# %%
# This code is just for illustration
from gpaw import GPAW, PW
images = [initial]
for i in range(n_im):
    ranks = range(i * n, (i + 1) * n)
    image = initial.copy()
    image.set_constraint(constraint)
    if world.rank in ranks:
        calc = GPAW(mode=PW(350),
                    nbands='130%',
                    xc='PBE',  # student: ...,
                    txt=f'{i}.txt',
                    communicator=ranks)
        image.calc = calc
    images.append(image)
images.append(final)

# %%
"""
When running the parallel NEB, you should choose the number of CPU cores
properly.  Let Ncore = N_im * Nk where N_im is the number of images, and Nk
is a divisor of the number of k-points; i.e. if there are 6 irreducible
k-point, Nk should be 1, 2, 3 or 6.  Keep the total number of cores to 24 or
less, or your job will wait too long in the queue.
"""

# %%
"""
### Input parameters

Some suitable parameters for the NEB are given below:

* Use the same calculator and constraints as for the initial and final images, but remember to set the `communicator` as described above
* Use 6 images. This gives a reasonable description of the energy landscape and can be run e.g. on 12 cores.
* Use a spring constant of 1.0 between the images. A lower value will slow the convergence
* Relax the initial NEB until `fmax=0.1eV/Å`, then switch on the climbing image and relax until `fmax=0.05eV/Å`.
"""

# %%
# teacher:
from gpaw import GPAW, PW
from ase.visualize import view
from ase.optimize import BFGS
from ase.parallel import world

initial = read('N2Ru.traj')
final = read('2Nads.traj')

images = [initial]
N = 4  # No. of images

z = initial.positions[:, 2]
constraint = FixAtoms(mask=(z < z.min() + 1.0))

j = world.rank * N // world.size
n = world.size // N  # number of cpu's per image

for i in range(N):
    ranks = range(i * n, (i + 1) * n)
    image = initial.copy()
    image.set_constraint(constraint)
    if world.rank in ranks:
        calc = GPAW(xc='PBE',
                    mode=PW(350),
                    nbands='130%',
                    communicator=ranks,
                    txt=f'{i}.txt',
                    kpts={'size': (4, 4, 1), 'gamma': True},
                    convergence={'eigenstates': 1e-7})
        image.calc = calc
    images.append(image)
images.append(final)

neb = NEB(images, k=1.0, parallel=True)
neb.interpolate()
qn = BFGS(neb, logfile='neb.log', trajectory='neb.traj')
qn.run(fmax=0.1)
neb.climb = True
qn.run(fmax=0.05)

# %%
"""
Once the calculation is done you should check that the final path looks
reasonable. What is the N-N distance at the saddle point? Use NEBTools to
calculate the barrier. Is N<sub>2</sub> likely to dissociate on the surface
at room temperature?
"""

# %%
# teacher:
# Check results:
images[1:5] = read('neb.traj@-5:-1')
energies = [image.get_potential_energy() for image in images]
emax = max(energies)
assert energies[2] == emax
assert abs(emax - energies[0] - 1.32) < 0.02
d = images[2].get_distance(-1, -2)
assert abs(d - 1.777) < 0.004



# web-page: E_w.png
import numpy as np
import matplotlib.pyplot as plt

A = np.loadtxt('frequency_equidistant.dat').transpose()
plt.plot(A[0], A[1], label='Equidistant')
B = np.loadtxt('frequency_gauss16.dat').transpose()
plt.plot(B[0], B[1], 'o', label='Gauss-Legendre 16')

plt.xlabel('Frequency [eV]', fontsize=18)
plt.ylabel('Integrand', fontsize=18)
plt.axis([0, 50, None, None])
plt.legend(loc='lower right')
plt.savefig('E_w.png')



from myqueue.workflow import run


def workflow():
    with run(script='Na2TDDFT.py', cores=2, tmax='1h'):
        run(script='part2.py')
    with run(script='ground_state.py', cores=8):
        run(script='spectrum.py')



# web-page: systems.db
from ase.optimize.test.test import all_optimizers
from ase.optimize.test.systems import create_database


def workflow():
    from myqueue.workflow import run

    with run(function=create_database):
        runs = [run(script='run_tests_emt.py')]

    for name in all_optimizers:
        if name == 'Berny':
            continue
        runs.append(run(script='run_tests.py',
                        args=[name], cores=8, tmax='1d'))

    run(script='analyze.py', deps=runs)



scalapack = True

libraries = ['scalapack-openmpi',
             'blacsCinit-openmpi',
             'blacsF77init-openmpi',
             'blacs-openmpi',
             'xc',
             'blas',
             'lapack']

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



from gpaw import GPAW, LCAO
from ase import Atoms
import numpy as np
from gpaw.directmin.etdm import ETDM
# Water molecule:
d = 0.9575
t = np.pi / 180 * 104.51
H2O = Atoms('OH2',
            positions=[(0, 0, 0),
                       (d, 0, 0),
                       (d * np.cos(t), d * np.sin(t), 0)])
H2O.center(vacuum=5.0)

calc = GPAW(mode=LCAO(),
            basis='dzp',
            eigensolver=ETDM(
                searchdir_algo={'name': 'l-bfgs-p', 'memory': 10}),
            occupations={'name': 'fixed-uniform'},
            mixer={'backend': 'no-mixing'},
            nbands='nao'
            )
H2O.set_calculator(calc)
H2O.get_potential_energy()



from ase.build import bulk
from gpaw import GPAW
from gpaw.raman.dipoletransition import get_momentum_transitions

atoms = bulk('C', 'diamond', a=3.567)
calc = GPAW(mode='lcao', basis='dzp',
            kpts=(5, 5, 5), xc='PBE',
            symmetry={'point_group': False},
            convergence={'bands': 'nao', 'density': 1e-5},
            parallel={'band': 1},
            txt='mom.txt')

atoms.calc = calc
atoms.get_potential_energy()
atoms.calc.write("gs.gpw", mode="all")
get_momentum_transitions(calc.wfs, savetofile=True)



"""Test point-group CLI."""
import pytest
from ase.io.cube import write_cube

from gpaw import GPAW
from gpaw.point_groups.cli import main


@pytest.mark.serial
def test_pg_cli(gpw_files, capsys):
    file = gpw_files['h2o_lcao_wfs']
    main(f'C2v {file} -c O -b 0:4'.split())
    out = capsys.readouterr().out
    lines = [line.split() for line in out.splitlines()]
    assert lines[1][1] == 'Yes'
    assert '-'.join(lines[n][4] for n in range(3, 7)) == 'A1-B2-A1-B1'


@pytest.mark.serial
def test_pg_cli_cube(gpw_files, capsys, in_tmp_dir):
    calc = GPAW(gpw_files['h2o_lcao_wfs'])
    wf = calc.get_pseudo_wave_function(0)
    cube = in_tmp_dir / 'h2o.cube'
    with cube.open('w') as fd:
        write_cube(fd, calc.atoms, wf)
    main(f'C2v {cube} -c O'.split())
    out = capsys.readouterr().out
    lines = [line.split() for line in out.splitlines()]
    assert lines[1][1] == 'Yes'
    assert lines[3][4] == 'A1'



"""This module calculates XC kernels for response function calculations.
"""

import numpy as np

from gpaw.response.kxc import AdiabaticSusceptibilityFXC
from gpaw.response.tms import get_goldstone_scaling


def get_xc_kernel(pd, chi0, functional='ALDA', kernel='density',
                  rshelmax=-1, rshewmin=None,
                  chi0_wGG=None,
                  fxc_scaling=None,
                  density_cut=None,
                  spinpol_cut=None):
    """
    Factory function that calls the relevant functions below
    """

    if kernel == 'density':
        return get_density_xc_kernel(pd, chi0, functional=functional,
                                     rshelmax=rshelmax, rshewmin=rshewmin,
                                     chi0_wGG=chi0_wGG,
                                     density_cut=density_cut)
    elif kernel in ['+-', '-+']:
        # Currently only collinear adiabatic xc kernels are implemented
        # for which the +- and -+ kernels are the same
        return get_transverse_xc_kernel(pd, chi0, functional=functional,
                                        rshelmax=rshelmax, rshewmin=rshewmin,
                                        chi0_wGG=chi0_wGG,
                                        fxc_scaling=fxc_scaling,
                                        density_cut=density_cut,
                                        spinpol_cut=spinpol_cut)
    else:
        raise ValueError('%s kernels not implemented' % kernel)


def get_density_xc_kernel(pd, chi0, functional='ALDA',
                          rshelmax=-1, rshewmin=None,
                          chi0_wGG=None,
                          density_cut=None):
    """
    Density-density xc kernels
    Factory function that calls the relevant functions below
    """

    calc = chi0.calc
    fd = chi0.fd
    nspins = len(chi0.gs.nt_sG)
    assert nspins == 1

    if functional[0] == 'A':
        # Standard adiabatic kernel
        print('Calculating %s kernel' % functional, file=fd)
        Kcalc = AdiabaticSusceptibilityFXC(calc, functional,
                                           world=chi0.world, txt=fd,
                                           timer=chi0.timer,
                                           rshelmax=rshelmax,
                                           rshewmin=rshewmin,
                                           density_cut=density_cut)
        Kxc_GG = Kcalc('00', pd)
        if pd.kd.gamma:
            Kxc_GG[0, :] = 0.0
            Kxc_GG[:, 0] = 0.0
        Kxc_sGG = np.array([Kxc_GG])
    elif functional[:2] == 'LR':
        print('Calculating LR kernel with alpha = %s' % functional[2:],
              file=fd)
        Kxc_sGG = calculate_lr_kernel(pd, calc, alpha=float(functional[2:]))
    elif functional == 'Bootstrap':
        print('Calculating Bootstrap kernel', file=fd)
        Kxc_sGG = get_bootstrap_kernel(pd, chi0, chi0_wGG, fd)
    else:
        raise ValueError('Invalid functional for the density-density '
                         'fxc kernel:', functional)

    return Kxc_sGG[0]


def get_transverse_xc_kernel(pd, chi0, functional='ALDA_x',
                             rshelmax=-1, rshewmin=None,
                             chi0_wGG=None,
                             fxc_scaling=None,
                             density_cut=None,
                             spinpol_cut=None):
    """ +-/-+ xc kernels
    Currently only collinear ALDA kernels are implemented
    Factory function that calls the relevant functions below
    """

    calc = chi0.calc
    fd = chi0.fd
    nspins = len(calc.density.nt_sG)
    assert nspins == 2

    if functional in ['ALDA_x', 'ALDA_X', 'ALDA']:
        # Adiabatic kernel
        print("Calculating transverse %s kernel" % functional, file=fd)
        Kcalc = AdiabaticSusceptibilityFXC(calc, functional,
                                           world=chi0.world, txt=fd,
                                           timer=chi0.timer,
                                           rshelmax=rshelmax,
                                           rshewmin=rshewmin,
                                           density_cut=density_cut,
                                           spinpol_cut=spinpol_cut)
    else:
        raise ValueError("%s spin kernel not implemented" % functional)

    Kxc_GG = Kcalc('+-', pd)

    if fxc_scaling is not None:
        assert isinstance(fxc_scaling[0], bool)
        if fxc_scaling[0]:
            if fxc_scaling[1] is None:
                assert pd.kd.gamma
                print('Finding rescaling of kernel to fulfill the '
                      'Goldstone theorem', file=chi0.fd)
                mode = fxc_scaling[2]
                assert mode in ['fm', 'afm']
                fxc_scaling[1] = get_goldstone_scaling(mode,
                                                       chi0.wd.omega_w,
                                                       -chi0_wGG, Kxc_GG,
                                                       world=chi0.world)

            assert isinstance(fxc_scaling[1], float)
            Kxc_GG *= fxc_scaling[1]

    return Kxc_GG


def calculate_lr_kernel(pd, calc, alpha=0.2):
    """Long range kernel: fxc = \alpha / |q+G|^2"""

    assert pd.kd.gamma

    f_G = np.zeros(len(pd.G2_qG[0]))
    f_G[0] = -alpha
    f_G[1:] = -alpha / pd.G2_qG[0][1:]

    return np.array([np.diag(f_G)])


def get_bootstrap_kernel(pd, chi0, chi0_wGG, fd):
    """ Bootstrap kernel (see below) """

    if chi0.world.rank == 0:
        chi0_GG = chi0_wGG[0]
        if chi0.world.size > 1:
            # If size == 1, chi0_GG is not contiguous, and broadcast()
            # will fail in debug mode.  So we skip it until someone
            # takes a closer look.
            chi0.world.broadcast(chi0_GG, 0)
    else:
        nG = pd.ngmax
        chi0_GG = np.zeros((nG, nG), complex)
        chi0.world.broadcast(chi0_GG, 0)

    return calculate_bootstrap_kernel(pd, chi0_GG, fd)


def calculate_bootstrap_kernel(pd, chi0_GG, fd):
    """Bootstrap kernel PRL 107, 186401"""

    if pd.kd.gamma:
        v_G = np.zeros(len(pd.G2_qG[0]))
        v_G[0] = 4 * np.pi
        v_G[1:] = 4 * np.pi / pd.G2_qG[0][1:]
    else:
        v_G = 4 * np.pi / pd.G2_qG[0]

    nG = len(v_G)
    K_GG = np.diag(v_G)

    fxc_GG = np.zeros((nG, nG), dtype=complex)
    dminv_GG = np.zeros((nG, nG), dtype=complex)

    for iscf in range(120):
        dminvold_GG = dminv_GG.copy()
        Kxc_GG = K_GG + fxc_GG

        chi_GG = np.dot(np.linalg.inv(np.eye(nG, nG)
                                      - np.dot(chi0_GG, Kxc_GG)), chi0_GG)
        dminv_GG = np.eye(nG, nG) + np.dot(K_GG, chi_GG)

        alpha = dminv_GG[0, 0] / (K_GG[0, 0] * chi0_GG[0, 0])
        fxc_GG = alpha * K_GG
        print(iscf, 'alpha =', alpha, file=fd)
        error = np.abs(dminvold_GG - dminv_GG).sum()
        if np.sum(error) < 0.1:
            print('Self consistent fxc finished in %d iterations !' % iscf,
                  file=fd)
            break
        if iscf > 100:
            print('Too many fxc scf steps !', file=fd)

    return np.array([fxc_GG])



from math import cos, pi, sin

from gpaw.new.ase_interface import GPAW
from myqueue.workflow import run


def workflow():
    with run(script='VCl2.py', cores=4):
        run(script='plot.py')
        run(function=check)


def check():
    calc = GPAW('VCl2_gs.gpw')
    M_v, M_av = calc.calculation.state.density.calculate_magnetic_moments()
    print(M_v)
    print(M_av)
    m = 2.50
    for x, m_v in zip([0, 2 * pi / 3, 4 * pi / 3], M_av):
        assert abs(m_v - [m * cos(x), m * sin(x), 0]).max() < 0.01


if __name__ == '__main__':
    check()



# %%
"""
# Discover new ferromagnetic monolayers with high critical temperatures

Now that you know how to calculate exchange coupling constants and single-ion anisotropies, you are in a position to search for new ferromagnetic monolayers yourself.

First, we will try to find some suitable candidates based on the data, which is already in the C2DB.

1.   Go to the C2DB website at https://cmrdb.fysik.dtu.dk/c2db and search for magnetic monolayers with finite band gaps (the theory developed so far is not applicable to metals, since these typically have long range interactions). (Hint: Toggle `Magnetic` to `yes` and set a minimum value for the `Band gap range` in the search menu, then press the search icon.)
2.   Refine the query to sort out the magnetic monolayers with a ferromagnetic ground state. (Hint: Key values such as the nearest neighbor exchange coupling can be queried by entering `J>0` into the search field.)
3.   Show the spin state, exchange coupling and magnetic anisotropy in the property overview. (Hint: You can add key values to the overview through the `Add Column` botton.)
4.   Decide on a material you find promising to have a high Curie temperature, based on what you learned in previous notebooks. You may also want to choose a material with a limited number of atoms in the unit cell in order for the calculations to run fast.

Secondly, you will write your own code below to

5.   Calculate the energy difference per magnetic site of a ferromagnetic and an antiferromagnetic configuration.
6.   Calculate $J$.
7.   Calculate the single-ion anisotropy $A$.
8.   Estimate the Curie temperature $T_c$.

You are welcome to repeat the process for as many monolayers as you like.
"""

# %%



from pathlib import Path
from ase.io import read
from gpaw import GPAW
from gpaw.point_groups import SymmetryChecker

charges = {'Ih': -2,
           'Ico': -2,
           'Th': 2}

for path in Path().glob('Oh-*.xyz'):
    print(path, end='')
    pg = path.name.split('-')[0]
    atoms = read(path)
    atoms.center(vacuum=5)
    if 1:
        atoms.calc = GPAW(mode='lcao',
                          charge=charges.get(pg, 0.0),
                          txt=path.with_suffix('.txt'))
        atoms.get_potential_energy()
        atoms.calc.write(path.with_suffix('.gpw'), mode='all')
    else:
        atoms.calc = GPAW(path.with_suffix('.gpw'))
    center = atoms.get_center_of_mass()
    R = atoms.positions
    if pg in {'Ico', 'Ih'}:
        z = R[17] - R[23]
        x = (R[5] + R[11]) / 2 - R[9]
    else:
        x = None
        z = None
    checker = SymmetryChecker(pg, center, 4.5,
                              x=x, z=z)
    ok = checker.check_atoms(atoms)
    assert ok

    checker.check_calculation(atoms.calc,
                              0, atoms.calc.get_number_of_bands(),
                              output=path.with_suffix('.sym'))

    for n in range(atoms.calc.get_number_of_bands()):
        result = checker.check_band(atoms.calc, n)
        characters = result['characters']
        best = result['symmetry']
        for sym, value in characters.items():
            if sym != best:
                assert abs(value) < 0.1 * characters[best]



from myqueue.workflow import run


def workflow():
    with run(script='gs_Si.py', cores=4, tmax='20m'):
        with run(script='eps_Si.py', cores=4, tmax='6h'):
            run(script='plot_Si.py')

    with run(script='gs_MoS2.py', cores=4, tmax='1h'):
        with run(script='pol_MoS2.py', cores=64, tmax='33h'):
            run(script='plot_MoS2.py')
        with run(script='get_2d_eps.py', tmax='8h'):
            run(script='plot_2d_eps.py')
        run(script='alpha_MoS2.py')



from ase.build import mx2
from gpaw import GPAW
from gpaw.lcao.scissors import Scissors

a12 = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.13,
          size=(1, 1, 1))
a12 += mx2(formula='WS2', kind='2H', a=3.184, thickness=3.15,
           size=(1, 1, 1))
a12.positions[3:, 2] += 3.6 + 3.184
a12.center(vacuum=3.0, axis=2)

k = 6
a12.calc = GPAW(mode='lcao',
                basis='dzp',
                nbands='nao',
                kpts=(k, k, 1),
                eigensolver=Scissors([(-0.5, 0.5, 3),
                                      (-0.3, 0.3, 3)]),
                txt='12.txt')
a12.get_potential_energy()
a12.calc.write('12.gpw')

bp = a12.cell.bandpath('GMKG', npoints=80)

c12 = a12.calc.fixed_density(kpts=bp, symmetry='off')
bs = c12.band_structure()
bs.write('12bs.json')



# creates: ug.csv, ugf.csv, pw.csv, pwe.csv, m.csv
from gpaw.core import UniformGrid, PlaneWaves
from gpaw.core.plane_waves import PlaneWaveExpansions
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.core.matrix import Matrix


for cls in [UniformGrid,
            PlaneWaves,
            UniformGridFunctions,
            PlaneWaveExpansions,
            Matrix]:
    name = ''.join(x for x in cls.__name__ if x.isupper()).lower()
    mod = cls.__module__
    if name == 'ug':
        mod = mod.replace('.uniform_grid', '')
    elif name == 'pw':
        mod = mod.replace('.plane_waves', '')
    print(name, mod)
    with open(f'{name}.csv', 'w') as fd:
        for name, meth in cls.__dict__.items():
            if name[0] != '_':
                try:
                    doc = meth.__doc__.splitlines()[0]
                except AttributeError:
                    doc = '...'
                print(f':meth:`~{mod}.{cls.__name__}.{name}`, "{doc}"',
                      file=fd)



import time
import pytest
import numpy as np

from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW, FermiDirac
from gpaw.test import findpeak, equal
from gpaw.response.df import DielectricFunction
from gpaw.response.susceptibility import FourComponentSusceptibilityTensor
from gpaw.response.susceptibility import read_macroscopic_component
from gpaw.mpi import size, world


@pytest.mark.kspair
@pytest.mark.response
def test_response_silicon_chi_RPA(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation

    t1 = time.time()

    a = 5.431
    atoms = bulk('Si', 'diamond', a=a)
    atoms.center()
    calc = GPAW(mode=PW(200),
                nbands=8,
                kpts=(4, 4, 4),
                parallel={'domain': 1},
                occupations=FermiDirac(width=0.05),
                xc='LDA')

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('Si', 'all')
    t2 = time.time()

    # Excited state calculation
    q = np.array([1 / 4.0, 0, 0])
    w = np.linspace(0, 24, 241)

    # Using DF
    df = DielectricFunction(calc='Si',
                            frequencies=w, eta=0.2, ecut=50,
                            hilbert=False)
    df.get_dynamic_susceptibility(xc='RPA', q_c=q, filename='Si_chi1.csv')

    t3 = time.time()

    world.barrier()

    # Using FCST
    fcst = FourComponentSusceptibilityTensor(calc, fxc='RPA',
                                             eta=0.2, ecut=50)
    fcst.get_macroscopic_component('00', q, w, filename='Si_chi2.csv')

    t4 = time.time()

    world.barrier()

    parprint('')
    parprint('For ground  state calc, it took', (t2 - t1) / 60, 'minutes')
    parprint('For excited state calc 1, it took', (t3 - t2) / 60, 'minutes')
    parprint('For excited state calc 2, it took', (t4 - t3) / 60, 'minutes')

    # The two response codes should hold identical results
    d1 = np.loadtxt('Si_chi1.csv', delimiter=',')
    wpeak1, Ipeak1 = findpeak(d1[:, 0], -d1[:, 4])
    w_w, chiks_w, chi_w = read_macroscopic_component('Si_chi2.csv')
    wpeak2, Ipeak2 = findpeak(w_w, -chi_w.imag)

    equal(wpeak1, wpeak2, 0.02)
    equal(Ipeak1, Ipeak2, 1.0)



import pytest

from ase.build import molecule

from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter
from gpaw.mpi import world, serial_comm
from gpaw.utilities import compiled_with_sl

from ..lcaotddft.test_molecule import only_on_master


pytestmark = pytest.mark.usefixtures('module_tmp_path')


def calculate_time_propagation(gpw_fpath, *,
                               iterations=3,
                               kick=[1e-5, 1e-5, 1e-5],
                               propagator='SICN',
                               communicator=world,
                               write_and_continue=False,
                               force_new_dm_file=False,
                               parallel={}):
    td_calc = TDDFT(gpw_fpath,
                    propagator=propagator,
                    communicator=communicator,
                    parallel=parallel,
                    txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat',
                       force_new_file=force_new_dm_file)
    if kick is not None:
        td_calc.absorption_kick(kick)
    td_calc.propagate(20, iterations)
    if write_and_continue:
        td_calc.write('td.gpw', mode='all')
        # Switch dipole moment writer and output
        td_calc.observers.pop()
        dm = DipoleMomentWriter(td_calc, 'dm2.dat', force_new_file=True)
        dm._update(td_calc)
        td_calc.propagate(20, iterations)
    communicator.barrier()


def check_dm(ref_fpath, fpath, rtol=1e-8, atol=1e-12):
    from gpaw.tddft.spectrum import read_dipole_moment_file

    world.barrier()
    _, time_ref_t, _, dm_ref_tv = read_dipole_moment_file(ref_fpath)
    _, time_t, _, dm_tv = read_dipole_moment_file(fpath)
    assert time_t == pytest.approx(time_ref_t, abs=0)
    assert dm_tv == pytest.approx(dm_ref_tv, rel=rtol, abs=atol)


# Generate different parallelization options
parallel_i = [{}]
if world.size > 1:
    parallel_i.append({'band': 2})
if compiled_with_sl():
    parallel_i.append({'sl_auto': True})
    if world.size > 1:
        parallel_i.append({'sl_auto': True, 'band': 2})


@pytest.fixture(scope='module')
@only_on_master(world)
def ground_state():
    atoms = molecule('SiH4')
    atoms.center(vacuum=4.0)

    calc = GPAW(nbands=6, h=0.4,
                convergence={'density': 1e-8},
                communicator=serial_comm,
                xc='LDA',
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')


@pytest.fixture(scope='module')
@only_on_master(world)
def time_propagation_reference(ground_state):
    calculate_time_propagation('gs.gpw',
                               communicator=serial_comm,
                               write_and_continue=True)


@pytest.mark.skip_for_new_gpaw
def test_dipole_moment_values(time_propagation_reference,
                              module_tmp_path, in_tmp_dir):
    with open('dm.dat', 'w') as fd:
        fd.write('''
# DipoleMomentWriter[version=1](center=False, density='comp')
#            time            norm                    dmx                    dmy                    dmz
          0.00000000       6.92701356e-16    -3.798602757097e-08    -3.850923113536e-10    -2.506988148420e-10
# Kick = [    1.000000000000e-05,     1.000000000000e-05,     1.000000000000e-05]; Time = 0.00000000
          0.00000000       6.78612525e-16    -3.806745480191e-08    -5.880500044945e-10    -4.683685533214e-10
          0.82682747      -3.11611967e-16     6.011432043389e-05     6.015251317290e-05     6.015179500177e-05
          1.65365493       1.71405522e-15     1.075009677567e-04     1.075385921602e-04     1.075337414463e-04
          2.48048240       1.55070479e-15     1.388363880650e-04     1.388662733804e-04     1.388701173331e-04
'''.strip())  # noqa: E501

    with open('dm2.dat', 'w') as fd:
        fd.write('''
# DipoleMomentWriter[version=1](center=False, density='comp')
#            time            norm                    dmx                    dmy                    dmz
          2.48048240       1.55070479e-15     1.388363880650e-04     1.388662733804e-04     1.388701173331e-04
          3.30730987      -1.85697397e-16     1.528174640313e-04     1.528352677280e-04     1.528428543052e-04
          4.13413733      -6.23799730e-17     1.497979692345e-04     1.498097215567e-04     1.498150038226e-04
          4.96096480       1.44537040e-15     1.324352983945e-04     1.324326482531e-04     1.324473352117e-04
'''.strip())  # noqa: E501

    rtol = 5e-4
    atol = 1e-8
    check_dm('dm.dat', module_tmp_path / 'dm.dat', rtol=rtol, atol=atol)
    check_dm('dm2.dat', module_tmp_path / 'dm2.dat', rtol=rtol, atol=atol)


@pytest.mark.skip_for_new_gpaw
@pytest.mark.parametrize('parallel', parallel_i)
@pytest.mark.parametrize('propagator', [
    'SICN', 'ECN', 'ETRSCN', 'SIKE'])
def test_propagation(time_propagation_reference,
                     parallel, propagator,
                     module_tmp_path, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'gs.gpw',
                               propagator=propagator,
                               parallel=parallel)
    atol = 1e-12
    if propagator == 'SICN':
        # This is the same propagator as the reference;
        # error comes only from parallelization
        rtol = 1e-8
        if 'band' in parallel:
            # XXX band parallelization is inaccurate!
            rtol = 7e-4
            atol = 3e-8
    else:
        # Other propagators match qualitatively
        rtol = 5e-2
        if 'band' in parallel:
            # XXX band parallelization is inaccurate!
            atol = 3e-8
    check_dm(module_tmp_path / 'dm.dat', 'dm.dat', rtol=rtol, atol=atol)


@pytest.mark.skip_for_new_gpaw
@pytest.mark.parametrize('parallel', parallel_i)
def test_restart(time_propagation_reference,
                 parallel,
                 module_tmp_path, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'td.gpw',
                               kick=None,
                               force_new_dm_file=True,
                               parallel=parallel)
    rtol = 1e-8
    if 'band' in parallel:
        rtol = 5e-4
    check_dm(module_tmp_path / 'dm2.dat', 'dm.dat', rtol=rtol)



from math import cos, sin, pi

from ase import Atoms
from ase.optimize import QuasiNewton
from ase.vibrations import Vibrations

from gpaw import GPAW

# Water molecule:
d = 0.9575
t = pi / 180 * 104.51

H2O = Atoms('H2O',
            positions=[(0, 0, 0),
                       (d, 0, 0),
                       (d * cos(t), d * sin(t), 0)])

H2O.center(vacuum=3.5)

calc = GPAW(h=0.2, txt='h2o.txt', mode='lcao', basis='dzp')

H2O.calc = calc

QuasiNewton(H2O).run(fmax=0.05)


"""Calculate the vibrational modes of a H2O molecule."""

# Create vibration calculator
vib = Vibrations(H2O)
vib.run()
vib.summary(method='frederiksen')

# Make trajectory files to visualize normal modes:
for mode in range(9):
    vib.write_mode(mode)



import os
import time

import numpy as np

from ase import Atoms
from ase.parallel import paropen
from ase.units import fs
from ase.io import Trajectory
from ase.md.verlet import VelocityVerlet
from gpaw import GPAW
from gpaw.mpi import world

name = 'na2_md'

# Equilibrium distance in Ang cf. setups page for Na dimer
d_bond = 3.29  # ~3.0 during oscillation
d_disp = 0.1

# Timestep and expected oscillatory period in attoseconds
timestep = 500.0
period = 2.1e5  # ~19.7 meV cf. CRC Handbook of Phys. & Chem. #09_08_91

ndiv = int(np.ceil(10e3 / timestep))  # update stats every 10 fs
niter = ndiv * int(np.ceil(3 * period / (ndiv * timestep)))


class Timing:
    def __init__(self, f):
        self.f = f
        self.t0 = time.time()
        self.i = 0

    def __call__(self, atoms):
        rate = 60 * ndiv / (time.time() - self.t0)
        ekin = atoms.get_kinetic_energy()
        epot = atoms.get_potential_energy()
        print('i=%06d (%6.2f min^-1), ekin=%13.9f, epot=%13.9f, etot=%13.9f' %
              (self.i, rate, ekin, epot, ekin + epot), file=self.f)
        self.t0 = time.time()
        self.i += ndiv


if __name__ == '__main__':
    if not os.path.isfile(name + '_gs.gpw'):
        atoms = Atoms('Na2', positions=[(0, 0, 0), (0, 0, d_bond + d_disp)])
        atoms.set_pbc(False)
        atoms.center(vacuum=6.0)
        atoms.set_velocities(np.zeros_like(atoms.get_positions()))
        cell_c = np.sum(atoms.get_cell()**2, axis=1)**0.5
        N_c = 16 * np.round(cell_c / (0.25 * 16))
        calc = GPAW(gpts=N_c, nbands=1, basis='dzp', setups={'Na': '1'},
                    txt=name + '_gs.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write(name + '_gs.gpw', mode='all')
        del atoms, calc
        time.sleep(10)

    while not os.path.isfile(name + '_gs.gpw'):
        print('Node %d waiting for file...' % world.rank)
        time.sleep(10)
    world.barrier()

    tdcalc = GPAW(name + '_gs.gpw', txt=name + '_td.txt')
    atoms = tdcalc.get_atoms()

    traj = Trajectory(name + '_td.traj', 'w', atoms)
    verlet = VelocityVerlet(atoms, timestep * 1e-3 * fs,
                            logfile=paropen(name + '_td.verlet', 'w'),
                            trajectory=traj)
    verlet.attach(Timing(paropen(name + '_td.log', 'w')), ndiv, atoms)
    verlet.run(niter)
    traj.close()



from myqueue.workflow import run


def workflow():
    Au111 = run(script='run-Au111.py', cores=24, tmax='1h')
    run(script='plot-traces.py', cores=1, tmax='1h', deps=[Au111])
    H_seq = run(script='run-Au111-H-seq.py', cores=24, tmax='5h', deps=[Au111])
    H_sim = run(script='run-Au111-H-sim.py', cores=24, tmax='5h', deps=[Au111])
    run(script='plot-seq-sim.py', cores=1, tmax='1h', deps=[H_seq, H_sim])
    H_hollow = run(script='run-Au111-H-hollow.py', cores=24, tmax='5h',
                   deps=[H_sim])
    neb = run(script='run-neb.py', cores=24, tmax='24h',
              deps=[H_sim, H_hollow])
    run(script='plot-neb.py', cores=1, tmax='1h', deps=[neb])



from myqueue.workflow import run
from gpaw.utilities import compiled_with_libvdwxc
from gpaw.xc.libvdwxc import libvdwxc_has_pfft


def workflow():
    run(script='hydrogen_atom.py', cores=16)
    if compiled_with_libvdwxc():
        run(script='libvdwxc-example.py')
        if libvdwxc_has_pfft():
            run(script='libvdwxc-pfft-example.py', cores=8)



# web-page: atomization.txt

from ase import Atoms
from ase.parallel import paropen as open
from gpaw import GPAW, PW, FermiDirac

a = 10.  # Size of unit cell (Angstrom)
c = a / 2

# Hydrogen atom:
atom = Atoms('H',
             positions=[(c, c, c)],
             magmoms=[0],
             cell=(a, a + 0.0001, a + 0.0002))  # Break cell symmetry

# gpaw calculator:
calc = GPAW(mode=PW(),
            xc='PBE',
            hund=True,
            eigensolver='rmm-diis',  # This solver can parallelize over bands
            occupations=FermiDirac(0.0, fixmagmom=True),
            txt='H.out',
            )
atom.calc = calc

e1 = atom.get_potential_energy()
calc.write('H.gpw')

# Hydrogen molecule:
d = 0.74  # Experimental bond length
molecule = Atoms('H2',
                 positions=([c - d / 2, c, c],
                            [c + d / 2, c, c]),
                 cell=(a, a, a))

calc.set(txt='H2.out')
calc.set(hund=False)  # No hund rule for molecules

molecule.calc = calc
e2 = molecule.get_potential_energy()
calc.write('H2.gpw')

with open('atomization.txt', 'w') as fd:
    print(f'  hydrogen atom energy:     {e1:5.2f} eV', file=fd)
    print(f'  hydrogen molecule energy: {e2:5.2f} eV', file=fd)
    print(f'  atomization energy:       {2 * e1 - e2:5.2f} eV', file=fd)



from ase.io import read
from gpaw import setup_paths
from gpaw import GPAW

# Insert the path to the created basis set
setup_paths.insert(0, '.')

atoms = read('r-methyloxirane.xyz')
atoms.center(vacuum=6.0)

calc = GPAW(mode='lcao',
            basis='aug.dzp',
            h=0.3,
            xc='PBE',
            nbands=16,
            poissonsolver={'remove_moment': 1 + 3 + 5},
            convergence={'density': 1e-12},
            txt='gs.out')
atoms.calc = calc
atoms.get_potential_energy()
calc.write('gs.gpw', mode='all')



from gpaw import setup_paths
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.magneticmomentwriter import MagneticMomentWriter

# Insert the path to the created basis set
setup_paths.insert(0, '.')


def main(kick):
    kick_strength = [0., 0., 0.]
    kick_strength['xyz'.index(kick)] = 1e-5

    td_calc = LCAOTDDFT('gs.gpw', txt=f'td-{kick}.out')

    DipoleMomentWriter(td_calc, f'dm-{kick}.dat')
    MagneticMomentWriter(td_calc, f'mm-{kick}.dat')

    td_calc.absorption_kick(kick_strength)
    td_calc.propagate(10, 2000)

    td_calc.write(f'td-{kick}.gpw', mode='all')


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--kick', default='x')
    kwargs = vars(parser.parse_args())
    main(**kwargs)



# web-page: Al_conv_ecut.png, Al_conv_k.png
import matplotlib.pyplot as plt

from ase.eos import EquationOfState
from ase.io import read


def fit(filename):
    configs = read(filename + '@:')
    volumes = [a.get_volume() for a in configs]
    energies = [a.get_potential_energy() for a in configs]
    eos = EquationOfState(volumes, energies)
    v0, e0, B = eos.fit()
    return (4 * v0)**(1 / 3.0)


cutoffs = range(200, 501, 50)
a = [fit(f'Al-{ecut}.txt') for ecut in cutoffs]
plt.figure(figsize=(6, 4))
plt.plot(cutoffs, a, 'o-')
plt.axis(ymin=4.03, ymax=4.05)
plt.xlabel('Plane-wave cutoff energy [eV]')
plt.ylabel('lattice constant [Ang]')
plt.savefig('Al_conv_ecut.png')

kpoints = range(4, 17)
plt.figure(figsize=(6, 4))
a = [fit(f'Al-{k:02}.txt') for k in kpoints]
plt.plot(kpoints, a, '-')
plt.xlabel('number of k-points')
plt.ylabel('lattice constant [Ang]')
plt.savefig('Al_conv_k.png')



from ase import Atoms
from gpaw import GPAW

name = 'diamond333_hch'

a = 3.574

atoms = Atoms('C8', [(0, 0, 0),
                     (1, 1, 1),
                     (2, 2, 0),
                     (3, 3, 1),
                     (2, 0, 2),
                     (0, 2, 2),
                     (3, 1, 3),
                     (1, 3, 3)],
              cell=(4, 4, 4),
              pbc=True)

atoms.set_cell((a, a, a), scale_atoms=True)
atoms *= (3, 3, 3)

calc = GPAW(h=0.2,
            txt=name + '.txt',
            xc='PBE',
            setups={0: 'hch1s'})

atoms.calc = calc

e = atoms.get_potential_energy()

calc.write(name + '.gpw')



from ase.visualize import view
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.build import fcc100, add_adsorbate
from gpaw import GPAW, PW

# Initial state:
# 2x2-Al(001) surface with 1 layer and an
# Au atom adsorbed in a hollow site:
slab = fcc100('Al', size=(2, 2, 2))
slab.center(axis=2, vacuum=3.0)
add_adsorbate(slab, 'Au', 1.6, 'hollow')

# Make sure the structure is correct:
view(slab)

# Fix the Al atoms:
mask = [atom.symbol == 'Al' for atom in slab]
print(mask)
fixlayer = FixAtoms(mask=mask)
slab.set_constraint(fixlayer)

# Use GPAW:
calc = GPAW(mode=PW(200), kpts=(2, 2, 1), xc='PBE', txt='hollow.txt')
slab.calc = calc

qn = QuasiNewton(slab, trajectory='hollow.traj')

# Find optimal height.  The stopping criterion is: the force on the
# Au atom should be less than 0.05 eV/Ang
qn.run(fmax=0.05)

calc.write('hollow.gpw')  # Write gpw output after the minimization

print('energy:', slab.get_potential_energy())
print('height:', slab.positions[-1, 2] - slab.positions[0, 2])



# creates: h2.emt.traj
from ase import Atoms
from ase.calculators.emt import EMT
from ase.optimize import QuasiNewton

system = Atoms('H2', positions=[[0.0, 0.0, 0.0],
                                [0.0, 0.0, 1.0]])
calc = EMT()

system.calc = calc

opt = QuasiNewton(system, trajectory='h2.emt.traj')

opt.run(fmax=0.05)



# web-page: e-spiral.png, m-spiral.png
import numpy as np
import matplotlib.pyplot as plt
from ase.spectrum.band_structure import BandStructure
from gpaw.new.ase_interface import GPAW

energies = []
magmoms = []
for i in range(31):
    atoms = GPAW(f'gs-{i:02}.gpw').get_atoms()
    energy = atoms.get_potential_energy()
    magmom = atoms.calc.calculation.magmoms()[0]
    energies.append(energy)
    magmoms.append(np.linalg.norm(magmom))

energies = np.array(energies) * 1000
energies -= energies[0]
magmoms = np.array(magmoms)

path = atoms.cell.bandpath('GXW', npoints=31)

bs = BandStructure(path, energies[np.newaxis, :, np.newaxis])
bs.plot(emin=-25, emax=5, label='q', ylabel='Energy [meV per atom]')
plt.savefig('e-spiral.png')

bs = BandStructure(path, magmoms[np.newaxis, :, np.newaxis])
bs.plot(emin=-0.1, emax=1.5, label='q',
        ylabel=r'Total magnetic moment [$\mu_B$]')
plt.savefig('m-spiral.png')



from gpaw.xc.fxc import FXCCorrelation
from ase.parallel import paropen

fxc = FXCCorrelation('diam_kern.ralda.lda_wfcs.gpw', xc='rALDA',
                     txt='diam_kern.ralda_03_ralda_wave.txt',
                     av_scheme='wavevector')
E_i = fxc.calculate(ecut=[131.072])

resultfile = paropen('diam_kern.ralda_kernel_comparison.dat', 'w')
resultfile.write(str(E_i[-1]) + '\n')
resultfile.close()



# web-page: sf_3x3_defect.png
from gpaw import GPAW
from gpaw.unfold import plot_spectral_function

calc = GPAW('gs_3x3_defect.gpw', txt=None)
ef = calc.get_fermi_level()

plot_spectral_function(filename='sf_3x3_defect',
                       color='blue',
                       eref=ef,
                       emin=-3,
                       emax=3)



import numpy as np
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.mom import prepare_mom_calculation
from gpaw.directmin.etdm import ETDM
from gpaw.directmin.tools import excite


@pytest.mark.mom
def test_mom_directopt_lcao_spinpaired(in_tmp_dir):
    atoms = Atoms('C2H4',
                  [[6.68748500e-01, 2.00680000e-04, 5.55800000e-05],
                   [-6.68748570e-01, -2.00860000e-04, -5.51500000e-05],
                   [4.48890600e-01, -5.30146300e-01, 9.32670330e-01],
                   [4.48878120e-01, -5.30176640e-01, -9.32674730e-01],
                   [-1.24289513e+00, 1.46164400e-02, 9.32559990e-01],
                   [-1.24286000e+00, -1.46832100e-02, -9.32554970e-01]])
    atoms.center(vacuum=4)

    eigensolver = ETDM(searchdir_algo={'name': 'l-sr1p'},
                       linesearch_algo={'name': 'max-step'})

    calc = GPAW(mode='lcao',
                basis='dzp',
                h=0.24,
                xc='PBE',
                symmetry='off',
                occupations={'name': 'fixed-uniform'},
                eigensolver=eigensolver,
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                convergence={'density': 1.0e-4,
                             'eigenstates': 4.0e-8})
    atoms.calc = calc
    atoms.get_potential_energy()

    f_sn = excite(calc, 0, 0, spin=(0, 0))
    f_sn[0] /= 2

    prepare_mom_calculation(calc, atoms, f_sn)
    # This fails if the memory of the search direction
    # algorithm is not erased
    e = atoms.get_potential_energy()

    calc.wfs.occupations.initialize_reference_orbitals()
    calc.wfs.calculate_occupation_numbers(calc.density.fixed)

    # These fail if the OccupationsMOM.numbers are not updated correctly
    assert np.all(calc.get_occupation_numbers() <= 2.0)
    assert e == pytest.approx(-21.38257404436053, abs=0.01)



from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw import PW
from gpaw.response.g0w0 import G0W0

a = 3.567
atoms = bulk('C', 'diamond', a=a)

calc = GPAW(mode=PW(300),
            parallel={'domain': 1},
            kpts=(3, 3, 3),
            xc='LDA',
            occupations=FermiDirac(0.001),
            txt='C_groundstate_freq.txt')

atoms.calc = calc
atoms.get_potential_energy()

calc.diagonalize_full_hamiltonian()
calc.write('C_groundstate_freq.gpw', 'all')

for i, domega0 in enumerate([0.01, 0.02, 0.03, 0.04, 0.05]):
    for j, omega2 in enumerate([1, 5, 10, 15, 20, 25]):
        frequencies = {'type': 'nonlinear',
                       'domega0': domega0,
                       'omega2': omega2}
        gw = G0W0(calc='C_groundstate_freq.gpw',
                  nbands=30,
                  bands=(3, 5),
                  kpts=[0],
                  ecut=20,
                  frequencies=frequencies,
                  filename=f'C_g0w0_domega0_{domega0}_omega2_{omega2}')

        results = gw.calculate()



# web-page: con_freq.png
import numpy as np
import matplotlib.pyplot as plt

A = np.loadtxt('con_freq.dat').transpose()
plt.plot(A[0], A[1], 'o-')

plt.xlabel('Number of frequency points', fontsize=18)
plt.ylabel('Energy', fontsize=18)
plt.axis([None, None, -6.7, -6.3])
plt.savefig('con_freq.png')



from ase import Atoms
from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter, RestartFileWriter
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField

# Na2 cluster
atoms = Atoms(symbols='Na2',
              positions=[(0, 0, 0), (3.0, 0, 0)],
              pbc=False)
atoms.center(vacuum=6.0)

# Standard ground state calculation
calc = GPAW(nbands=2, h=0.4, setups={'Na': '1'})
atoms.calc = calc
energy = atoms.get_potential_energy()
calc.write('na2_gs.gpw', mode='all')

# Standard time-propagation initialization
time_step = 10.0
iterations = 3000
kick_strength = [1.0e-3, 0.0, 0.0]
td_calc = TDDFT('na2_gs.gpw')
DipoleMomentWriter(td_calc, 'na2_td_dm.dat')
RestartFileWriter(td_calc, 'na2_td.gpw')

# Create and attach InducedField object
frequencies = [1.0, 2.08]     # Frequencies of interest in eV
folding = 'Gauss'             # Folding function
width = 0.1                   # Line width for folding in eV
ind = TDDFTInducedField(paw=td_calc,
                        frequencies=frequencies,
                        folding=folding,
                        width=width,
                        restart_file='na2_td.ind')

# Propagate as usual
td_calc.absorption_kick(kick_strength=kick_strength)
td_calc.propagate(time_step, iterations)

# Save TDDFT and InducedField objects
td_calc.write('na2_td.gpw', mode='all')
ind.write('na2_td.ind')



"""Calculate the Heisenberg exchange constants in Fe and Co using the MFT.
Test with unrealisticly loose parameters to catch if the numerics change.
"""

# General modules
import pytest
import numpy as np

# Script modules
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw import mpi
from gpaw.response.chiks import ChiKS
from gpaw.response.mft import IsotropicExchangeCalculator
from gpaw.response.site_kernels import (SphericalSiteKernels,
                                        CylindricalSiteKernels,
                                        ParallelepipedicSiteKernels)
from gpaw.response.heisenberg import (calculate_single_site_magnon_energies,
                                      calculate_fm_magnon_energies)


@pytest.mark.response
def test_Fe_bcc(in_tmp_dir):
    # ---------- Inputs ---------- #

    # Part 1: Ground state calculation
    xc = 'LDA'
    kpts = 4
    nbands = 6
    pw = 200
    occw = 0.01
    conv = {'density': 1e-8,
            'forces': 1e-8,
            'bands': nbands}
    a = 2.867
    mm = 2.21

    # Part 2: MFT calculation
    ecut = 50
    eta = 0.
    # Do the high symmetry points of the bcc lattice
    q_qc = np.array([[0, 0, 0],           # Gamma
                     [0.5, -0.5, 0.5],    # H
                     [0.0, 0.0, 0.5],     # N
                     ])
    # Define site kernels to test
    # Test a single site of spherical and cylindrical geometries
    rc_pa = np.array([[1.0], [1.5], [2.0]])
    hc_pa = np.array([[1.0], [1.5], [2.0]])
    ez_pav = np.array([[[1., 0., 0.]], [[0., 1., 0.]], [[0., 0., 1.]]])

    # ---------- Script ---------- #

    # Part 1: Ground state calculation

    atoms = bulk('Fe', 'bcc', a=a)
    atoms.set_initial_magnetic_moments([mm])
    atoms.center()

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts={'size': (kpts, kpts, kpts), 'gamma': True},
                nbands=nbands + 4,
                occupations=FermiDirac(occw),
                parallel={'domain': 1},
                spinpol=True,
                convergence=conv
                )

    atoms.calc = calc
    atoms.get_potential_energy()

    # Part 2: MFT calculation

    # Set up site kernels with a single site
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)
    sitekernels.append(CylindricalSiteKernels(positions, ez_pav,
                                              rc_pa, hc_pa))
    # Set up a kernel to fill out the entire unit cell
    sitekernels.append(ParallelepipedicSiteKernels(positions,
                                                   [[atoms.get_cell()]]))

    # Initialize the exchange calculator
    chiks = ChiKS(calc,
                  ecut=ecut, nbands=nbands, eta=eta,
                  gammacentered=True)
    isoexch_calc = IsotropicExchangeCalculator(chiks)

    # Allocate array for the exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Calcualate the exchange constant for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc(q_c, sitekernels)
    # Since we only have a single site, reduce the array
    J_qp = J_qabp[:, 0, 0, :]

    # Calculate the magnon energies
    mm_ap = mm * np.ones((1, npartitions))  # Magnetic moments
    mw_qp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)[:, 0, :]

    # Part 3: Compare results to test values
    test_J_pq = np.array([[1.61655323, 0.88149124, 1.10008928],
                          [1.86800734, 0.93735081, 1.23108285],
                          [4.67979867, 0.2004699, 1.28510023],
                          [1.14516166, 0.62140228, 0.78470217],
                          [1.734752, 0.87124284, 1.13880145],
                          [3.82381708, 0.31159032, 1.18094396],
                          [1.79888576, 0.92972442, 1.2054906]])
    test_mw_pq = np.array([[0., 0.66521177, 0.46738581],
                           [0., 0.84222041, 0.57640002],
                           [0., 4.05369028, 3.07212255],
                           [0., 0.47398746, 0.32620549],
                           [0., 0.78145439, 0.53931965],
                           [0., 3.17848551, 2.39173871],
                           [0., 0.78656857, 0.5370069]])

    # Exchange constants
    assert np.allclose(J_qp.imag, 0.)
    assert np.allclose(J_qp.real, test_J_pq.T, rtol=1e-3)
    
    # Magnon energies
    assert np.allclose(mw_qp, test_mw_pq.T, rtol=1e-3)


@pytest.mark.response
@pytest.mark.skipif(mpi.size == 1, reason='Slow test, skip in serial')
def test_Co_hcp(in_tmp_dir):
    # ---------- Inputs ---------- #

    # Part 1: Ground state calculation
    # Atomic configuration
    a = 2.5071
    c = 4.0695
    mm = 1.6
    # Ground state parameters
    xc = 'LDA'
    kpts = 6
    occw = 0.01
    nbands = 2 * (6 + 0)  # 4s + 3d + 0 empty shell bands
    ebands = 2 * 2  # extra bands for ground state calculation
    pw = 200
    conv = {'density': 1e-8,
            'forces': 1e-8,
            'bands': nbands}

    # Part 2: MFT calculation
    ecut = 100
    eta0 = 0.
    eta1 = 0.1
    # Do high symmetry points of the hcp lattice
    q_qc = np.array([[0, 0, 0],              # Gamma
                     [0.5, 0., 0.],          # M
                     [1. / 3., 1 / 3., 0.],  # K
                     [0., 0., 0.5]           # A
                     ])

    # Use spherical site kernels in a radius range which should yield
    # stable results
    rc_pa = np.array([[1.0, 1.0], [1.1, 1.1], [1.2, 1.2]])

    # Part 3: Compare results to test values
    # Unfortunately, the usage of symmetry leads to such extensive repetition
    # of random noise, that one cannot trust individual values of J very well.
    # This is improved when increasing the number of k-points, but the problem
    # never completely vanishes
    J_atol = 5.e-3
    J_rtol = 5.e-2
    # However, derived physical values have an increased error cancellation due
    # to their collective nature.
    mw_rtol = 5.e-3  # relative tolerance of absolute results
    mw_ctol = 5.e-2  # relative tolerance on kernel and eta self-consistency

    # ---------- Script ---------- #

    # Part 1: Ground state calculation

    atoms = bulk('Co', 'hcp', a=a, c=c)
    atoms.set_initial_magnetic_moments([mm, mm])
    atoms.center()

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts={'size': (kpts, kpts, kpts), 'gamma': True},
                occupations=FermiDirac(occw),
                convergence=conv,
                nbands=nbands + ebands,
                parallel={'domain': 1})

    atoms.calc = calc
    atoms.get_potential_energy()

    # Part 2: MFT calculation

    # Set up spherical site kernels
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)

    # Set up a site kernel to fill out the entire unit cell
    cell_cv = atoms.get_cell()
    cc_v = np.sum(cell_cv, axis=0) / 2.  # Unit cell center
    ucsitekernels = ParallelepipedicSiteKernels([cc_v], [[cell_cv]])

    # Initialize the exchange calculator with and without eta,
    # as well as with and without symmetry
    chiks0 = ChiKS(calc,
                   disable_point_group=True,
                   disable_time_reversal=True,
                   ecut=ecut, nbands=nbands, eta=eta0,
                   gammacentered=True)
    isoexch_calc0 = IsotropicExchangeCalculator(chiks0)
    chiks1 = ChiKS(calc,
                   ecut=ecut, nbands=nbands, eta=eta1,
                   gammacentered=True)
    isoexch_calc1 = IsotropicExchangeCalculator(chiks1)

    # Allocate array for the spherical site exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Allocate array for the unit cell site exchange constants
    Juc_qe = np.empty((nq, 2), dtype=complex)

    # Calcualate the exchange constants for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc0(q_c, sitekernels)
        Juc_qe[q, 0] = isoexch_calc0(q_c, ucsitekernels)[0, 0, 0]
        Juc_qe[q, 1] = isoexch_calc1(q_c, ucsitekernels)[0, 0, 0]

    # Calculate the magnon energy
    mm_ap = calc.get_magnetic_moment() / 2.\
        * np.ones((nsites, npartitions))
    mw_qnp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)
    mw_qnp = np.sort(mw_qnp, axis=1)  # Make sure the eigenvalues are sorted
    mwuc_qe = calculate_single_site_magnon_energies(Juc_qe, q_qc,
                                                    calc.get_magnetic_moment())

    # Part 3: Compare results to test values
    test_J_qab = np.array([[[1.37280875 - 0.j,
                             0.28516328 - 0.00007259j],
                            [0.28516328 + 0.00007259j,
                             1.37280875 - 0.j]],
                           [[0.99644998 + 0.j,
                             0.08202191 - 0.04867163j],
                            [0.08202191 + 0.04867163j,
                             0.99644998 + 0.j]],
                           [[0.95005156 - 0.j,
                             -0.03339854 - 0.05672191j],
                            [-0.03339854 + 0.05672191j,
                             0.950051561 + 0.j]],
                           [[1.30187481 - 0.j,
                             0.00000039 - 0.00525360j],
                            [0.00000039 + 0.00525360j,
                             1.30187481 + 0.j]]])
    test_J_qab = np.array([[[1.37280847 + 0.j, 0.28516320 + 0.00007375j],
                            [0.28516320 - 0.00007375j, 1.37280847 - 0.j]],
                           [[0.99649489 - 0.j, 0.08201540 - 0.04905246j],
                            [0.08201540 + 0.04905246j, 0.99649489 + 0.j]],
                           [[0.95009010 + 0.j, -0.0329297 - 0.05777656j],
                            [-0.0329297 + 0.05777656j, 0.95009010 + 0.j]],
                           [[1.30186322 - 0.j, 0.00000038 - 0.00478552j],
                            [0.00000038 + 0.00478552j, 1.30186322 - 0.j]]])
    test_mw_qn = np.array([[0., 0.673172311],
                           [0.667961643, 0.893557698],
                           [0.757038564, 0.914026524],
                           [0.414677028, 0.425972649]])
    test_mwuc_q = np.array([0., 0.72440073, 1.2123005, 0.37567975])

    # Exchange constants
    # err = np.absolute(J_qabp[..., 1] - test_J_qab)
    # is_bad = err > J_atol + J_rtol * np.absolute(test_J_qab)
    # print(is_bad)
    # print(np.absolute(err[is_bad] / np.absolute(test_J_qab[is_bad])))
    assert np.allclose(J_qabp[..., 1], test_J_qab,
                       atol=J_atol, rtol=J_rtol)

    # Magnon energies
    assert np.all(np.abs(mw_qnp[0, 0, :]) < 1.e-8)  # Goldstone theorem
    assert np.allclose(mwuc_qe[0, :], 0.)  # Goldstone
    assert np.allclose(mw_qnp[1:, 0, 1], test_mw_qn[1:, 0], rtol=mw_rtol)
    assert np.allclose(mw_qnp[:, 1, 1], test_mw_qn[:, 1], rtol=mw_rtol)
    assert np.allclose(mwuc_qe[1:, 0], test_mwuc_q[1:], rtol=mw_rtol)

    # Part 4: Check self-consistency of results
    # We should be in a radius range, where the magnon energies don't change
    assert np.allclose(mw_qnp[1:, 0, ::2], test_mw_qn[1:, 0, np.newaxis],
                       rtol=mw_ctol)
    assert np.allclose(mw_qnp[:, 1, ::2], test_mw_qn[:, 1, np.newaxis],
                       rtol=mw_ctol)
    # Check that a finite eta does not change the magnon energies too much
    assert np.allclose(mwuc_qe[1:, 0], mwuc_qe[1:, 1], rtol=mw_ctol)



import json
from ase.parallel import paropen
from ase.build import bulk
from ase.optimize import BFGS
from gpaw import GPAW, PW
from gpaw.zero_field_splitting import zfs


def nv_minus(n: int, relax: bool = False) -> None:
    atoms = bulk('C', 'diamond', cubic=True) * n
    atoms.numbers[0] = 7
    del atoms[1]
    atoms.set_initial_magnetic_moments([1] * len(atoms))
    name = f'NC{8 * n**3 - 2}'
    atoms.calc = GPAW(xc='PBE',
                      mode=PW(ecut=400),
                      charge=-1,
                      txt=name + '.txt')
    atoms.get_forces()
    # atoms.calc.write(name + '.gpw', 'all')
    D1 = zfs(atoms.calc)

    if relax:
        BFGS(atoms,
             logfile=name + '.log').run(fmax=0.05)
        # atoms.calc.write(name + '.relaxed.gpw', 'all')

    D2 = zfs(atoms.calc)
    with paropen(name + '.json', 'w') as fd:
        json.dump(dict(D1=D1.tolist(), D2=D2.tolist()), fd)


if __name__ == '__main__':
    for n in range(2, 4):
        nv_minus(n, relax=True)



# web-page: H2O_ir_summary.txt
from ase import io
from ase.vibrations.infrared import Infrared

atoms = io.read('relaxed.traj')
ir = Infrared(atoms, name='ir')
with open('H2O_ir_summary.txt', 'w') as fd:
    ir.summary(log=fd)



from ase import Atoms
from gpaw.calculator import GPAW as OldGPAW
from gpaw.new.ase_interface import GPAW as NewGPAW


def test_refactored_code(in_tmp_dir):
    new('n')


def new(x):
    params = {'mode': {'name': 'fd', 'force_complex_dtype': 0},
              'random': not True,
              'kpts': (4, 1, 1),
              'spinpol': not True,
              'convergence': {'maximum iterations': 200}}

    if x == 'n':
        GPAW = NewGPAW
    else:
        GPAW = OldGPAW

    atoms = Atoms('H2', cell=[2, 2, 3], pbc=True)
    atoms.positions[1, 2] = 0.8
    atoms.calc = GPAW(**params,
                      txt=f'{x}s.txt')
    f = atoms.get_forces()
    e = atoms.get_potential_energy()
    atoms.get_dipole_moment()
    print(e)
    print(f)
    if 1:
        atoms.positions[1, 2] = 0.75
        f2 = atoms.get_forces()
        e2 = atoms.get_potential_energy()
        print(f2)
        print(e2)


if __name__ == '__main__':
    import sys
    for x in sys.argv[1:]:
        new(x)



import matplotlib.pyplot as plt
from ase import Atoms
from ase.units import Bohr
from gpaw.utilities.ps2ae import PS2AE
from gpaw import GPAW

plt.rcParams['font.size'] = 12

hli = Atoms('HLi', positions=[[0, 0, 0], [0, 0, 1.6]])
hli.center(vacuum=2.5)
hli.calc = GPAW(txt='hli.txt', mode='fd')
hli.get_potential_energy()

# Transformer:
t = PS2AE(hli.calc, grid_spacing=0.05)

for n in range(2):
    ps = t.get_wave_function(n, ae=False)
    ae = t.get_wave_function(n)
    norm_squared = t.gd.integrate(ae**2) * Bohr**3
    print('Norm:', norm_squared)
    assert abs(norm_squared - 1) < 1e-2
    i = ps.shape[0] // 2
    x = t.gd.coords(2) * Bohr

    # Interpolated PS and AE wfs:
    plt.plot(x, ps[i, i], '--', color=f"C{n}",
             label=r'$\tilde\psi_{}$'.format(n))
    plt.plot(x, ae[i, i], '-', color=f"C{n}",
             label=r'$\psi_{}$'.format(n))

    # Raw PS wfs:
    ps0 = hli.calc.get_pseudo_wave_function(n, pad=True)
    gd = hli.calc.wfs.gd
    i = ps0.shape[0] // 2
    X = gd.coords(2) * Bohr
    plt.plot(X, ps0[i, i], 'o', color=f"C{n}")

plt.hlines(0, xmin=x[0], xmax=x[-1])
plt.xlabel(r'z [$\rm \AA$]')
plt.ylabel(r'wave functions [$\rm \AA^{-3/2}$]')
plt.legend()
plt.tight_layout()
plt.savefig('hli-wfs.png')
hli.calc.write('hli.gpw')
plt.close()



from myqueue.workflow import run


def workflow():
    run(script='bulk.py', cores=4)
    with run(script='surface.py', cores=4):
        with run(script='sigma.py'):
            run(script='fig2.py')



from gpaw.tddft import TDDFT
from gpaw.tddft.ehrenfest import EhrenfestVelocityVerlet
from gpaw.tddft.laser import CWField
from ase.units import Hartree, Bohr, AUT
from ase.io import Trajectory
from ase.parallel import parprint

name = 'h2_diss'

# Ehrenfest simulation parameters
timestep = 10.0  # timestep given in attoseconds
ndiv = 10  # write trajectory every 10 timesteps
niter = 500  # run for 500 timesteps

# TDDFT calculator with an external potential emulating an intense
# harmonic laser field aligned (CWField uses by default the z axis)
# along the H2 molecular axis.
tdcalc = TDDFT(name + '_gs.gpw',
               txt=name + '_td.txt',
               propagator='EFSICN',
               solver='BiCGStab',
               td_potential=CWField(1000 * Hartree, 1 * AUT, 10))

# For Ehrenfest dynamics, we use this object for the Velocity Verlet dynamics.
ehrenfest = EhrenfestVelocityVerlet(tdcalc)

# Trajectory to save the dynamics.
traj = Trajectory(name + '_td.traj', 'w', tdcalc.get_atoms())

# Propagates the dynamics for niter timesteps.
for i in range(1, niter + 1):
    ehrenfest.propagate(timestep)

    if tdcalc.atoms.get_distance(0, 1) > 2.0:
        # Stop simulation if H-H distance is greater than 2 A
        parprint('Dissociated!')
        break

    # Every ndiv timesteps, save an image in the trajectory file.
    if i % ndiv == 0:
        # Currently needed with Ehrenfest dynamics to save energy,
        # forces and velocitites.
        epot = tdcalc.get_td_energy() * Hartree
        F_av = ehrenfest.F * Hartree / Bohr  # forces
        v_av = ehrenfest.v * Bohr / AUT  # velocities
        atoms = tdcalc.atoms.copy()
        # Needed to save the velocities to the trajectory:
        atoms.set_velocities(v_av)

        traj.write(atoms, energy=epot, forces=F_av)

traj.close()



# web-page: Fe_magnon_dispersion.png
"""Plot the magnon dispersion of Fe(bcc)"""

# General modules
import numpy as np
from matplotlib import rcParams
import matplotlib.pyplot as plt

# Script modules
from gpaw import GPAW
from gpaw.response.heisenberg import calculate_single_site_magnon_energies

# ---------- Inputs ---------- #

# Ground state
gpw = 'Fe_all.gpw'

# High-symmetry points
sp_p = [r'$\Gamma$', 'N', 'P', 'H']
spq_pc = [np.array(qc) for qc in
          [[0., 0., 0], [0., 0., 0.5], [0.25, 0.25, 0.25], [0.5, -0.5, 0.5]]]

# Load MFT data
q_qc = np.load('Fe_q_qc.npy')
rc_r = np.load('Fe_rc_r.npy')
J_qr = np.load('Fe_J_qr.npy')
Juc_q = np.load('Fe_Juc_q.npy')

# Define range of spherical radii to plot
rmin = 1.0
rmax = 1.5

# Labels and limits
mwlabel = r'$\hbar\omega$ [meV]'
mwlim = (0., 450.)

filename = 'Fe_magnon_dispersion.png'

# ---------- Script ---------- #

# Extract the magnetization of the unit cell
calc = GPAW(gpw, txt=None)
muc = calc.get_magnetic_moment()

# Convert relative q-points into distance along the bandpath in reciprocal
# space.
B_cv = 2.0 * np.pi * calc.atoms.cell.reciprocal()  # Coordinate transform
q_qv = q_qc @ B_cv  # Transform into absolute reciprocal coordinates
pathq_q = [0.]
for q in range(1, len(q_qc)):
    pathq_q.append(pathq_q[-1] + np.linalg.norm(q_qc[q] - q_qc[q - 1]))
pathq_q = np.array(pathq_q)

# Define q-limits of plot
qlim = ((pathq_q[0] - pathq_q[1]) / 2.,
        (1.5 * pathq_q[-1] - 0.5 * pathq_q[-2]))

# Calculate the magnon energies
E_qr = calculate_single_site_magnon_energies(J_qr, q_qc, muc) * 1.e3  # meV
Euc_q = calculate_single_site_magnon_energies(Juc_q, q_qc, muc) * 1.e3

# We define the lower bound on the magnon energy as the minimum within the
# chosen rc range, the "best estimate" as the median, and the upper bound
# as the maximum value
Evalid_qr = E_qr[:, np.logical_and(rc_r >= rmin, rc_r <= rmax)]
Emin_q = np.min(Evalid_qr, axis=1)
E_q = np.median(Evalid_qr, axis=1)
Emax_q = np.max(Evalid_qr, axis=1)

# Plot the magnon dispersion with spherical sites
colors = rcParams['axes.prop_cycle'].by_key()['color']
plt.fill_between(pathq_q, Emin_q, Emax_q, color=colors[0], alpha=0.4)
plt.plot(pathq_q, Emin_q, color='0.5')
plt.plot(pathq_q, Emax_q, color='0.5')
plt.plot(pathq_q, E_q, color=colors[0], label='sphere')

# Plot the magnon dispersion with parallelepipedic sites
plt.plot(pathq_q, Euc_q, '-o', mec='k', color=colors[1], label='unit cell')

# Use high-symmetry points as tickmarks for the x-axis
qticks = []
qticklabels = []
for sp, spq_c in zip(sp_p, spq_pc):
    for q, q_c in enumerate(q_qc):
        if np.allclose(q_c, spq_c):
            qticks.append(pathq_q[q])
            qticklabels.append(sp)
plt.xticks(qticks, qticklabels)
# Plot also vertical lines for each special point
for pq in qticks:
    plt.axvline(pq, color='0.5', linewidth=1, zorder=0)

# Labels and limits
plt.ylabel(mwlabel)
plt.xlim(qlim)
plt.ylim(mwlim)

plt.legend(title='Site geometry')

plt.savefig(filename, format=filename.split('.')[-1],
            bbox_inches='tight')



import numpy as np
from gpaw.response.bse import BSE
from gpaw.response.df import DielectricFunction

ecut = 50
eshift = 0.8
eta = 0.05

df = DielectricFunction('gs_MoS2.gpw',
                        ecut=ecut,
                        frequencies=np.linspace(0, 5, 1001),
                        nbands=50,
                        intraband=False,
                        hilbert=False,
                        eta=eta,
                        eshift=eshift,
                        txt='rpa_MoS2.txt')

df.get_polarizability(filename='pol_rpa_MoS2.csv')

bse = BSE('gs_MoS2.gpw',
          spinors=True,
          ecut=ecut,
          valence_bands=[8],
          conduction_bands=[9],
          nbands=50,
          eshift=eshift,
          mode='BSE',
          txt='bse_MoS2.txt')

bse.get_polarizability(filename='pol_bse_MoS2.csv',
                       eta=eta,
                       write_eig='bse_MoS2_eig.dat',
                       w_w=np.linspace(0, 5, 5001))

bse = BSE('gs_MoS2.gpw',
          spinors=True,
          ecut=ecut,
          valence_bands=[8],
          conduction_bands=[9],
          truncation='2D',
          nbands=50,
          eshift=eshift,
          mode='BSE',
          txt='bse_MoS2_trun.txt')

bse.get_polarizability(filename='pol_bse_MoS2_trun.csv',
                       eta=eta,
                       write_eig='bse_MoS2_eig_trun.dat',
                       w_w=np.linspace(0, 5, 5001))



import numpy as np
from gpaw.mpi import world


def mpi_debug(data, ordered=True):
    global msgcount

    if not isinstance(data, list):
        data = [data]

    if ordered:
        for i in range(world.rank):
            world.barrier()

    for txt in data:
        print('%02d-mpi%d, %s' % (msgcount, world.rank, txt))
        if ordered:
            msgcount += 1

    if ordered:
        for i in range(world.size - world.rank):
            world.barrier()


W = world.size
N = 32

assert N % W == 0
M = N // W

# Create my share of data
data = np.arange(world.rank * M, (world.rank + 1) * M)

# Let's calculate the global sum
slocal = data.sum()
s = world.sum(slocal)
mpi_debug('data: %s, slocal=%d, s=%d' % (data, slocal, s))
assert s == N * (N - 1) // 2

# Subtract the global mean
data -= s / N
mpi_debug(f'data: {data}')

# -------------------------------------------------------------------

if world.rank == 0:
    print('-' * 16)

# Who has global index 11? The master needs it!
i = 11
rank, ilocal = divmod(i, M)
mpi_debug('rank=%d, ilocal=%d, i=%d' % (rank, ilocal, i))
assert rank * M + ilocal == i

# Do I have it?
if world.rank == rank:
    # Yes, so extract data (must be an array)
    idata = np.array([data[ilocal]], dtype=data.dtype)
else:
    # No, so just allocate space
    idata = np.empty(1, dtype=data.dtype)

# Broadcast from owner to everyone else
world.broadcast(idata, rank)

"""
# This does the same as broadcast with send/receive...

# Do I have it?
if world.rank == rank:
    # Yes, now send it to the others
    for other_rank in range(world.size):
        # We don't have to send it to ourselves
        if other_rank != rank:
            world.send(idata, other_rank, tag=123)
else:
    # No, so receive from the one that own the data
    world.receive(idata, rank, tag=123)
"""

mpi_debug('idata=%d' % idata)

# -------------------------------------------------------------------

if world.rank == 0:
    print('-' * 16)

# The master just calculated auxiliary data. Distribute it.
aux = np.empty(N, dtype=float)

# Only master knows the data right now
if world.rank == 0:
    np.random.seed(1234567)
    aux[:] = np.random.uniform(0, 1, size=N).round(2)
    print(f'MASTER aux: {aux}, mean={aux.mean():f}')

# Allocate space for my part of the auxiliary data
myaux = np.empty(M, dtype=float)

# Scatter parts from master to everyone
world.scatter(aux, myaux, 0)

"""
# This does the same as scatter with send/receive...

# Are we the master?
if world.rank == 0:
    # Yes, so extract my part directly
    myaux[:] = aux[0:M]

    # Now send parts to the slaves
    for slave_rank in range(1, world.size):
        youraux = aux[slave_rank*M:(slave_rank+1)*M]
        world.send(youraux, slave_rank, tag=123)
else:
    # No, so receive from the master
    world.receive(myaux, 0, tag=123)
"""

# We don't need original data anymore
del aux

# Try to calculate mean now
meanaux = world.sum(myaux.mean()) / world.size
mpi_debug(f'myaux: {myaux}, mean={meanaux:f}')

# -------------------------------------------------------------------

if world.rank == 0:
    print('-' * 16)

# We've done something to our part of the auxiliary data. Master needs it all
if world.rank == 0:
    result = np.empty(N, dtype=float)
else:
    result = None

# Do something to our auxiliary data
myaux[:] = np.sin(2 * np.pi * myaux).round(3)
mpi_debug(f'myaux: {myaux}')

# Gather parts from everyone on the master
world.gather(myaux, 0, result)

"""
# This does the same as gather with send/receive...

# Are we the master?
if world.rank == 0:
    # Yes, so extract my part directly
    result[0:M] = myaux[:]

    # Now receive parts from the slaves
    for slave_rank in range(1, world.size):
        youraux = np.empty(M, dtype=float)
        world.receive(youraux, slave_rank, tag=123)
        result[slave_rank*M:(slave_rank+1)*M] = youraux
else:
    # No, so send to the master
    world.send(myaux, 0, tag=123)
"""

mpi_debug(f'result: {result}')



# web-page: Ag55_spec_basis.png
import numpy as np
import matplotlib.pyplot as plt

data_ej = np.loadtxt('spec.dat')
data_my_ej = np.loadtxt('mybasis/spec.dat')
data_dzp_ej = np.loadtxt('dzp/spec.dat')

plt.figure(figsize=(6, 6 / 1.62))
ax = plt.subplot(1, 1, 1)
ax.plot(data_ej[:, 0], data_ej[:, 1], 'k', label='p-valence')
ax.plot(data_my_ej[:, 0], data_my_ej[:, 1], 'k--', label='p-valence (my)')
ax.plot(data_dzp_ej[:, 0], data_dzp_ej[:, 1], 'k:', label='dzp')
ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
ax.yaxis.set_ticks_position('left')
ax.xaxis.set_ticks_position('bottom')
plt.title(r'Effect of LCAO basis on the spectrum')
plt.xlabel('Energy (eV)')
plt.legend(loc='upper left', title='LCAO basis')
plt.ylabel('Photoabsorption (eV$^{-1}$)')
plt.xlim(0, 6)
plt.ylim(ymin=0)
plt.tight_layout()
plt.savefig('Ag55_spec_basis.png')



"""MoS2 slab.

Parallelization is done over:

* k-points (embarrassingly parallel)
* electronic bands
* domain

and we must have K*B*D=world.size.

Problem size can be tweaked by adjusting N and ecut.

The calculation has several sections:

* importing modules
* first iteration (contains initialization stuff)
* more iterations (should all be similar)
* forces
* stress tensor

"""
from time import time
from ase.build import mx2
from gpaw import GPAW
from gpaw.mpi import world

K = 8
B = 1
D = world.size // (K * B)
assert K * D * B == world.size

N = 6
atoms = mx2('MoS2')
atoms.center(vacuum=4.0, axis=2)
atoms *= (N, N, 1)
atoms.rattle(stdev=0.01)

# from ase.visualize import view
# view(atoms)

ecut = 600
atoms.calc = GPAW(
    kpts=(4, 4, 1),
    xc='PBE',
    txt=f'MoS2-{N}x{N}-k{K}-d{D}-b{B}.txt',
    parallel={
        'domain': D,
        'band': B,
        'kpt': K,
        'sl_auto': True},
    mode={'name': 'pw', 'ecut': ecut})

t = time()
for iter, _ in enumerate(atoms.calc.icalculate(atoms)):
    if world.rank == 0:
        print(iter, time() - t)
        t = time()
    if iter == 4:
        atoms.calc.scf.converged = True
        break

atoms.get_forces()
atoms.get_stress()



from gpaw import GPAW, PW, FermiDirac
from ase.build import bulk
si = bulk('Si')
si.calc = GPAW(mode=PW(200),
               parallel={'domain': 1},
               kpts={'size': (2, 2, 2), 'gamma': True},
               occupations=FermiDirac(0.01))
si.get_potential_energy()
si.calc.diagonalize_full_hamiltonian()
si.calc.write('Si_gs.gpw', 'all')
dct = {}
exec(open('Si_g0w0_ppa.py').read(), dct)
assert abs(dct['ks_gap'] - 0.404) < 0.01
assert abs(dct['qp_gap'] - 1.138) < 0.01



import pytest
from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.mpi import serial_comm
from gpaw.xc.rpa import RPACorrelation


@pytest.mark.response
def test_rpa_rpa_energy_Si(in_tmp_dir):
    a0 = 5.43
    Si = bulk('Si', a=a0)

    calc = GPAW(mode='pw',
                kpts={'size': (2, 2, 2), 'gamma': True},
                occupations=FermiDirac(0.001),
                communicator=serial_comm)
    Si.calc = calc
    Si.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=50)

    ecut = 50
    rpa = RPACorrelation(calc, qsym=False, nfrequencies=8)
    E_rpa_noqsym = rpa.calculate(ecut=[ecut])

    rpa = RPACorrelation(calc, qsym=True, nfrequencies=8)
    E_rpa_qsym = rpa.calculate(ecut=[ecut])

    assert E_rpa_qsym == pytest.approx(-12.61, abs=0.01)
    assert E_rpa_qsym == pytest.approx(E_rpa_noqsym)



import pytest
from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
from gpaw import GPAW, FermiDirac
from gpaw.response.g0w0 import G0W0


@pytest.mark.response
def test_response_gw_MoS2_cut(in_tmp_dir, scalapack):
    if 1:
        calc = GPAW(mode='pw',
                    xc='PBE',
                    experimental={'niter_fixdensity': 2},
                    nbands=16,
                    setups={'Mo': '6'},
                    eigensolver='rmm-diis',
                    occupations=FermiDirac(0.001),
                    kpts={'size': (6, 6, 1), 'gamma': True})

        a = 3.1604
        c = 10.0

        cell = Hexagonal(symbol='Mo',
                         latticeconstant={'a': a, 'c': c}).get_cell()
        layer = Atoms(symbols='MoS2', cell=cell, pbc=True,
                      scaled_positions=[(0, 0, 0),
                                        (2 / 3, 1 / 3, 0.3),
                                        (2 / 3, 1 / 3, -0.3)])

        pos = layer.get_positions()
        pos[1][2] = pos[0][2] + 3.172 / 2
        pos[2][2] = pos[0][2] - 3.172 / 2
        layer.set_positions(pos)
        layer.calc = calc
        layer.get_potential_energy()
        calc.write('MoS2.gpw', mode='all')

    gw = G0W0('MoS2.gpw',
              'gw-test',
              nbands=15,
              ecut=10,
              eta=0.2,
              frequencies={'type': 'nonlinear', 'domega0': 0.1},
              truncation='2D',
              kpts=[((1 / 3, 1 / 3, 0))],
              bands=(8, 10),
              savepckl=True)

    e_qp = gw.calculate()['qp'][0, 0]

    ev = 2.669
    ec = 6.831
    assert e_qp[0] == pytest.approx(ev, abs=0.01)
    assert e_qp[1] == pytest.approx(ec, abs=0.01)



from numpy import reshape, dot

from ase.visualize import view
from ase.build import fcc111, add_adsorbate
from gpaw import GPAW
from gpaw.mixer import MixerSum
import gpaw.dscf as dscf

filename = 'lumo'

c_mol = GPAW(nbands=9,
             h=0.2,
             xc='RPBE',
             kpts=(8, 6, 1),
             spinpol=True,
             convergence={'energy': 100,
                          'density': 100,
                          'eigenstates': 1.0e-9,
                          'bands': -2},
             txt='CO_lumo.txt')

calc = GPAW(nbands=80,
            h=0.2,
            xc='RPBE',
            kpts=(8, 6, 1),
            eigensolver='cg',
            spinpol=True,
            mixer=MixerSum(nmaxold=5, beta=0.1, weight=100),
            convergence={'energy': 100,
                         'density': 100,
                         'eigenstates': 1.0e-7,
                         'bands': -10},
            txt=filename + '.txt')

# Import Slab with relaxed CO
slab = fcc111('Pt', size=(1, 2, 3), orthogonal=True)
add_adsorbate(slab, 'C', 2.0, 'ontop')
add_adsorbate(slab, 'O', 3.15, 'ontop')
slab.center(axis=2, vacuum=4.0)

view(slab)

molecule = slab.copy()

del molecule[:-2]

# Molecule
molecule.calc = c_mol
molecule.get_potential_energy()

# Find band corresponding to lumo
lumo = c_mol.get_pseudo_wave_function(band=5, kpt=0, spin=1)
lumo = reshape(lumo, -1)

wf1_k = [c_mol.get_pseudo_wave_function(band=5, kpt=k, spin=1)
         for k in range(c_mol.wfs.kd.nibzkpts)]
wf2_k = [c_mol.get_pseudo_wave_function(band=6, kpt=k, spin=1)
         for k in range(c_mol.wfs.kd.nibzkpts)]

band_k = []
for k in range(c_mol.wfs.kd.nibzkpts):

    wf1 = reshape(wf1_k[k], -1)
    wf2 = reshape(wf2_k[k], -1)
    p1 = abs(dot(wf1, lumo))
    p2 = abs(dot(wf2, lumo))
    if p1 > p2:
        band_k.append(5)
    else:
        band_k.append(6)

# Lumo wavefunction
wf_u = [kpt.psit_nG[band_k[kpt.k]] for kpt in c_mol.wfs.kpt_u]

# Lumo projector overlaps
mol = range(len(slab))[-2:]
p_uai = [dict([(mol[a], P_ni[band_k[kpt.k]]) for a, P_ni in kpt.P_ani.items()])
         for kpt in c_mol.wfs.kpt_u]

#   Slab with adsorbed molecule
slab.calc = calc
orbital = dscf.AEOrbital(calc, wf_u, p_uai)
dscf.dscf_calculation(calc, [[1.0, orbital, 1]], slab)
slab.get_potential_energy()



from ase import Atoms
from gpaw import GPAW, PW

a = 2.87
m = 2.2

fe = Atoms('Fe2',
           scaled_positions=[(0, 0, 0),
                             (0.5, 0.5, 0.5)],
           magmoms=[m, -m],
           cell=(a, a, a),
           pbc=True)

calc = GPAW(mode=PW(350),
            kpts=(6, 6, 6),
            txt='anti.txt')

fe.calc = calc
e = fe.get_potential_energy()
calc.write('anti.gpw')



# web-page: shg.png
import numpy as np
import matplotlib.pyplot as plt
from ase.io import read

# Plot and save both spectra
atoms = read('gs.txt')
cell = atoms.get_cell()
cellsize = atoms.get_cell_lengths_and_angles()
mult = cellsize[2] * 1e-10  # make the sheet sus.
legls = []
res_name = ['shg_yyy_lg.npy', 'shg_yyy_vg.npy']
plt.figure(figsize=(6.0, 4.0), dpi=300)
for ii, name in enumerate(res_name):
    # Load the data
    shg = np.load(name)
    w_l = shg[0]
    plt.plot(np.real(w_l), np.real(mult * shg[1] * 1e18), '-')
    plt.plot(np.real(w_l), np.imag(mult * shg[1] * 1e18), '--')
    legls.append(f'{name}: Re')
    legls.append(f'{name}: Im')
    plt.xlabel(r'$\hbar\omega$ (eV)')
    plt.ylabel(r'$\chi_yyy$ (nm$^2$/V)')
    plt.legend(legls, ncol=2)
plt.tight_layout()
plt.savefig('shg.png', dpi=300)



import asr


@asr.workflow
class MyWorkflow:
    atoms = asr.var()
    calculator = asr.var()

    @asr.task
    def relax(self):
        return asr.node('relax',
                        atoms=self.atoms,
                        calculator=self.calculator)
    # --- end-snippet-1 ---

    @asr.task
    def groundstate(self):
        return asr.node('groundstate', atoms=self.relax,
                        calculator=self.calculator)

    @asr.task
    def bandstructure(self):
        return asr.node('bandstructure', gpw=self.groundstate)


def workflow(runner):
    from ase.build import bulk
    wf = MyWorkflow(
        atoms=bulk('Si'),
        calculator={'mode': 'pw',
                    'kpts': (4, 4, 4),
                    'txt': 'gpaw.txt'})
    runner.run_workflow(wf)



import numpy as np
import pytest
from pytest import approx

from gpaw.point_groups import (PointGroup, SymmetryChecker,
                               point_group_names as names)


@pytest.mark.serial
@pytest.mark.parametrize('name', names)
def test_for_various_errors(name):
    pg = PointGroup(name)
    print(pg)
    pg.get_normalized_table()


@pytest.mark.serial
@pytest.mark.parametrize('name', names)
def test_translations(name):
    pg = PointGroup(name)
    # Calculate the symmetry representations of p-orbitals:
    x = np.linspace(-5, 5, 11)
    xyz = np.array(np.meshgrid(x, x, x, indexing='ij'))
    r2 = (xyz**2).sum(0)
    checker = SymmetryChecker(pg, center=(5, 5, 5), grid_spacing=1.5)
    for i in range(3):
        func = np.exp(-r2) * xyz[i]
        dct = checker.check_function(func, np.eye(3))
        print(i, dct)
        assert dct['symmetry'] == pg.translations[i]


@pytest.mark.serial
@pytest.mark.parametrize('name', names)
def test_pg(name):
    pg = PointGroup(name)

    # Check that the number of irreducible representation
    # is equal to the number of symmetry
    # transform classes:
    assert len(pg.character_table) == len(pg.character_table[0])

    h = sum(pg.nops)

    # Checks for groups with real character tables:
    if not pg.complex:
        # Check that the sum of squared dimensions of the irreps
        # equals to the number of symmetry elements h
        assert (pg.character_table[:, 0]**2).sum() == h

        # Rows:
        for i, row1 in enumerate(pg.character_table):

            # Check normalization:
            assert (row1**2).dot(pg.nops) == approx(h, 1e-8)

            for j, row2 in enumerate(pg.character_table):

                if i >= j:
                    continue

                # Check orthogonality:
                assert (row1 * row2).dot(pg.nops) == approx(0, abs=1e-8)

        # Columns:
        for i, row1 in enumerate(pg.character_table.T):

            # Check normalization:
            assert (row1**2).sum() == approx(h / pg.nops[i], abs=1e-13)

            for j, row2 in enumerate(pg.character_table.T):

                if i >= j:
                    continue

                # Check orthogonality:
                assert row1.dot(row2) == approx(0, 1e-8)

    # Checks for complex groups:
    else:
        reps = pg.symmetries
        # Rows:
        for i, row1 in enumerate(pg.character_table):

            # Check normalization:
            norm = (row1**2).dot(pg.nops)

            # Real rows:
            if reps[i].find('E') < 0:
                correctnorm = h
            else:  # complex rows
                correctnorm = h * 2

            assert norm == approx(correctnorm, abs=1e-8)

            for j, row2 in enumerate(pg.character_table):

                if i >= j:
                    continue

                # Compare real with real rows and complex with complex rows:
                if ((reps[i].find('E') >= 0 and reps[j].find('E') >= 0)
                    or
                    (reps[i].find('E') < 0 and reps[j].find('E') < 0)):

                    # Check orthogonality:
                    norm = (row1 * row2).dot(pg.nops)
                    assert norm == approx(0, abs=1e-13)
